class Activities:
    """A class to access the Activities PDI API endpoint."""

    def __init__(self):
        activites_endpoint = "/activities"
        self.url_activites = self.base_url + activites_endpoint
        super().__init__()

    def get_activities(self, limit: int = None):
        """Get a list of Activities.
        `Args:`
            limit: int
                Specify limit to return.

        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        return self._request(self.url_activites, limit=limit)

    def create_activity(self, activity_name: str, canvassing_shift: bool):
        """
        Create a New Activity
        `Args:`
            activity_name str: The activity name
            canvassing_shift bool: The canvassing shift
        """
        payload = {"activityName": activity_name, "canvassingShift": canvassing_shift}
        return self._request(self.url_activites, req_type="POST", post_data=payload)

    def get_activity(self, id: str):
        """
        Get a Activity by id.
        `Args:`
            id: str
                The Activity id
        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        return self._request(f"{self.url_activites}/{id}")

    def update_activity(self, id: str, activity_name: str, canvassing_shift: str):
        """
        Update an Activity
        `Args:`
            id: Activity id
            activity_name str: The activity name
            canvassing_shift bool: The canvassing shift
        """
        payload = {"activityName": activity_name, "canvassingShift": canvassing_shift}
        return self._request(f"{self.url_activites}/{id}", req_type="PUT", post_data=payload)



import os

import pytest

from parsons.pdi.pdi import PDI


@pytest.fixture
def live_pdi():
    # Generate a live PDI connection based on these env vars

    username = os.environ["PDI_USERNAME"]
    password = os.environ["PDI_PASSWORD"]
    api_token = os.environ["PDI_API_TOKEN"]

    pdi = PDI(username, password, api_token, qa_url=True)

    return pdi


@pytest.fixture
def mock_pdi(requests_mock):
    # Not meant to hit live api servers

    requests_mock.post(
        "https://apiqa.bluevote.com/sessions",
        json={
            "AccessToken": "AccessToken",
            "ExpirationDate": "2100-01-01",
        },
    )

    username = "PDI_USERNAME"
    password = "PDI_PASSWORD"
    api_token = "PDI_API_TOKEN"

    pdi = PDI(username, password, api_token, qa_url=True)

    return pdi



import os
from typing import Optional


def check(env: str, field: Optional[str], optional: Optional[bool] = False) -> Optional[str]:
    """
    Check if an environment variable has been set. If it has not been set
    and the passed field or arguments have not been passed, then raise an
    error.
    """
    if field:
        return field
    try:
        return os.environ[env]
    except KeyError as e:
        if not optional:
            raise KeyError(
                f"No {env} found. Store as environment variable or pass as an argument."
            ) from e



import os
import unittest

from parsons import SFTP, TargetSmartAutomation
from test.utils import mark_live_test


class TestTargetSmartAutomation(unittest.TestCase):
    def setUp(self):
        self.ts = TargetSmartAutomation()
        self.job_name = "a-test-job"
        self.sftp = SFTP(
            self.ts.sftp_host,
            os.environ["TS_SFTP_USERNAME"],
            os.environ["TS_SFTP_PASSWORD"],
            self.ts.sftp_port,
        )
        self.test_xml = "test/test_targetsmart/job_config.xml"

    def tearDown(self):
        # Clean up the files were put on the SFTP
        self.ts.remove_files(self.job_name)

    @mark_live_test
    def test_create_job_xml(self):
        # Assert that job xml creates the file correctly
        job_xml = self.ts.create_job_xml(
            "job_type", "match_job", ["test@gmail.com", "test2@gmail.com"]
        )
        with open(self.test_xml, "r") as xml:
            test_xml = xml.read()
        with open(job_xml, "r") as xml:
            real_xml = xml.read()
        self.assertEqual(test_xml, real_xml)

    @mark_live_test
    def test_config_status(self):
        # Find good configuration
        self.sftp.put_file(self.test_xml, f"{self.ts.sftp_dir}/{self.job_name}.job.xml.good")
        self.assertTrue(self.ts.config_status(self.job_name))
        self.ts.remove_files(self.job_name)

        # Find bad configuration
        self.sftp.put_file(self.test_xml, f"{self.ts.sftp_dir}/{self.job_name}.job.xml.bad")
        self.assertRaises(ValueError, self.ts.config_status, self.job_name)

    @mark_live_test
    def test_match_status(self):
        # Find good configuration
        good_match = "test/test_targetsmart/match_good.xml"
        self.sftp.put_file(good_match, f"{self.ts.sftp_dir}/{self.job_name}.finish.xml")
        self.assertTrue(self.ts.match_status(self.job_name))
        self.ts.remove_files(self.job_name)

        # Find bad configuration
        bad_match = "test/test_targetsmart/match_bad.xml"
        self.sftp.put_file(bad_match, f"{self.ts.sftp_dir}/{self.job_name}.finish.xml")
        self.assertRaises(ValueError, self.ts.match_status, self.job_name)

    @mark_live_test
    def test_remove_files(self):
        # Add a file
        self.sftp.put_file(self.test_xml, f"{self.ts.sftp_dir}/{self.job_name}.txt")

        # Remove files
        self.ts.remove_files(self.job_name)

        # Check that file is not there
        dir_list = self.sftp.list_directory(f"{self.ts.sftp_dir}/")
        self.assertNotIn(f"{self.job_name}.txt", dir_list)



import os
import unittest

from parsons import FacebookAds, Table

users_table = Table(
    [
        {
            "first": "Bob",
            "middle": "J",
            "last": "Smith",
            "phone": "1234567890",
            "cell": None,
            "vb_voterbase_dob": "19820413",
        },
        {
            "first": "Sue",
            "middle": "Lucy",
            "last": "Doe",
            "phone": None,
            "cell": "2345678901",
            "vb_voterbase_dob": None,
        },
    ]
)


@unittest.skipIf(not os.environ.get("LIVE_TEST"), "Skipping because not running live test")
class TestFacebookAdsIntegration(unittest.TestCase):
    def setUp(self):
        self.fb_ads = FacebookAds()

        self.audience_id = self.fb_ads.create_custom_audience(
            name="Test Audience", data_source="USER_PROVIDED_ONLY"
        )

    def tearDown(self):
        self.fb_ads.delete_custom_audience(self.audience_id)

    def test_create_custom_audience(self):
        # Audience created in setUp
        self.assertIsNotNone(self.audience_id)

    def test_create_custom_audience_bad_data_source(self):
        self.assertRaises(
            KeyError,
            self.fb_ads.create_custom_audience,
            name="Something",
            data_source="INVALID",
        )

    def test_add_users_to_custom_audience(self):
        # Note we don't actually check the results of adding these users, eg. how many were
        # matched by FB. This test just ensures the method doesn't error.
        # TODO See if we can do a more effective test.
        self.fb_ads.add_users_to_custom_audience(self.audience_id, users_table)

    def test_add_users_to_custom_audience_no_valid_columns(self):
        # We don't yet support full names for matching, so this shouldn't work
        tbl = Table(
            [
                {"full name": "Bob Smith"},
            ]
        )
        self.assertRaises(KeyError, self.fb_ads.add_users_to_custom_audience, self.audience_id, tbl)


class TestFacebookAdsUtilities(unittest.TestCase):
    def test_get_match_key_for_column(self):
        # Test just a few of the mappings
        self.assertEqual("EMAIL", FacebookAds._get_match_key_for_column("email"))
        self.assertEqual("EMAIL", FacebookAds._get_match_key_for_column("voterbase_email"))
        self.assertEqual("FN", FacebookAds._get_match_key_for_column("first name"))
        self.assertEqual("FN", FacebookAds._get_match_key_for_column("FIRST-NAME "))
        self.assertEqual("FN", FacebookAds._get_match_key_for_column("vb_tsmart_first_name"))
        self.assertEqual("LN", FacebookAds._get_match_key_for_column("Last Name!"))
        self.assertEqual("ST", FacebookAds._get_match_key_for_column("state code"))
        self.assertEqual("ST", FacebookAds._get_match_key_for_column("vb_vf_source_state"))
        self.assertEqual("GEN", FacebookAds._get_match_key_for_column("vb_voterbase_gender"))
        self.assertEqual(
            "PHONE",
            FacebookAds._get_match_key_for_column("vb_voterbase_phone_wireless"),
        )
        self.assertIsNone(FacebookAds._get_match_key_for_column("invalid"))

    def test_get_preprocess_key_for_column(self):
        self.assertEqual(
            "DOB YYYYMMDD",
            FacebookAds._get_preprocess_key_for_column("vb_voterbase_dob"),
        )

    def test_get_match_table_for_users_table(self):
        # This tests basic column matching, as well as the more complex cases like:
        # - Where multiple columns map to the same FB key ("PHONE").
        # - Parsing of a YYYYMMDD birth date column.
        t = FacebookAds.get_match_table_for_users_table(users_table)

        row0 = t[0]
        self.assertEqual(["FN", "LN", "PHONE", "DOBY", "DOBM", "DOBD"], t.columns)
        self.assertEqual("Bob", row0["FN"])
        self.assertEqual("Smith", row0["LN"])
        self.assertEqual("1234567890", row0["PHONE"])
        self.assertEqual("1982", row0["DOBY"])
        self.assertEqual("04", row0["DOBM"])
        self.assertEqual("13", row0["DOBD"])

        row1 = t[1]
        self.assertEqual("Sue", row1["FN"])
        self.assertEqual("Doe", row1["LN"])
        self.assertEqual("2345678901", row1["PHONE"])
        self.assertEqual("", row1["DOBY"])
        self.assertEqual("", row1["DOBM"])
        self.assertEqual("", row1["DOBD"])

    def test_get_match_schema_and_data(self):
        match_table = Table(
            [
                {"FN": "Bob", "LN": "Smith"},
                {"FN": "Sue", "LN": "Doe"},
            ]
        )
        (schema, data) = FacebookAds._get_match_schema_and_data(match_table)
        self.assertEqual(["FN", "LN"], schema)
        self.assertEqual(("Bob", "Smith"), data[0])
        self.assertEqual(("Sue", "Doe"), data[1])



import logging

from twilio.rest import Client

from parsons.etl import Table
from parsons.utilities import check_env, json_format

logger = logging.getLogger(__name__)


class Twilio:
    """
    Instantiate the Twilio class

    `Args:`
        account_sid: str
            The Twilio account sid. Not required if ``TWILIO_ACCOUNT_SID`` env variable is
            passed.
        auth_token: str
            The Twilio auth token. Not required if ``TWILIO_AUTH_TOKEN`` env variable is
            passed.
    `Returns`:
        Twilio class
    """

    def __init__(self, account_sid=None, auth_token=None):
        self.account_sid = check_env.check("TWILIO_ACCOUNT_SID", account_sid)
        self.auth_token = check_env.check("TWILIO_AUTH_TOKEN", auth_token)
        self.client = Client(self.account_sid, self.auth_token)

    def _table_convert(self, obj):
        tbl = Table([x.__dict__["_properties"] for x in obj])

        if "subresource_uris" in tbl.columns and "uri" in tbl.columns:
            tbl.remove_column("subresource_uris", "uri")

        return tbl

    def get_account(self, account_sid):
        """
        Get Twilio account

        `Args:`
            account_sid: str
                The Twilio account sid
        `Returns:`
            dict
        """

        r = self.client.api.accounts(account_sid)
        logger.info(f"Retrieved {account_sid} account.")
        return r.__dict__

    def get_accounts(self, name=None, status=None):
        """
        Get Twilio accounts including subaccounts.

        `Args:`
            name: str
                Filter to name of the account
            status: str
                Filter to an account status of ``active``, ``closed`` or ``suspended``.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self.client.api.accounts.list(friendly_name=name, status=status)
        tbl = self._table_convert(r)

        logger.info(f"Retrieved {tbl.num_rows} accounts.")
        return tbl

    def get_account_usage(
        self,
        category=None,
        start_date=None,
        end_date=None,
        time_period=None,
        group_by=None,
        exclude_null=False,
    ):
        """
        Get Twilio account usage.

        `Args:`
            category: str
                Filter to a specific type of usage category. The list of possibilities can be found
                `here <https://www.twilio.com/docs/usage/api/usage-record?code-sample=code-last-months-usage-for-all-usage-categories-4&code-language=Python&code-sdk-version=5.x#usage-all-categories>`_.
            start_date: str
                Filter to usage from a specific start date (ex. ``2019-01-01``).
            end_date: str
                Filter to usage from a specific end date (ex. ``2019-01-01``).
            time_period: str
                A convenience method to filter usage. Can be one of ``today``, ``yesterday``,
                ``this_month``, ``last_month``.
            group_by: str
                The time interval to group usage by. Can be one of ``daily``, ``monthly`` or
                ``yearly``.
            exclude_null: boolean
                Exclude rows that have no usage.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        # Add populated arguments
        args = {"category": category, "start_date": start_date, "end_date": end_date}
        args = json_format.remove_empty_keys(args)

        # Parse out the time_periods
        if time_period == "today":
            r = self.client.usage.records.today.list(**args)
        elif time_period == "yesterday":
            r = self.client.usage.records.yesterday.list(**args)
        elif time_period == "this_month":
            r = self.client.usage.records.this_month.list(**args)
        elif time_period == "last_month":
            r = self.client.usage.records.last_month.list(**args)

        # Parse out the group by
        elif group_by == "daily":
            r = self.client.usage.records.daily.list(**args)
        elif group_by == "monthly":
            r = self.client.usage.records.monthly.list(**args)
        elif group_by == "yearly":
            r = self.client.usage.records.yearly.list(**args)
        else:
            r = self.client.usage.records.list(**args)

        tbl = self._table_convert(r)

        if exclude_null:
            tbl.remove_null_rows("count", null_value="0")

        return tbl

    def get_messages(
        self,
        to=None,
        from_=None,
        date_sent=None,
        date_sent_before=None,
        date_sent_after=None,
    ):
        """
        Get Twilio messages.

        `Args:`
            to: str
                Filter to messages only sent to the specified phone number.
            from_: str
                Filter to messages only sent from the specified phone number.
            date_sent: str
                Filter to messages only sent on the specified date (ex. ``2019-01-01``).
            date_sent_before: str
                Filter to messages only sent before the specified date (ex. ``2019-01-01``).
            date_sent_after: str
                Filter to messages only sent after the specified date (ex. ``2019-01-01``).
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self.client.messages.list(
            to=to,
            from_=from_,
            date_sent=date_sent,
            date_sent_before=date_sent_before,
            date_sent_after=date_sent_after,
        )

        tbl = self._table_convert(r)
        logger.info(f"Retrieved {tbl.num_rows} messages.")
        return tbl



import unittest

import requests_mock

from parsons import Table, Zoom
from test.utils import assert_matching_tables

ACCOUNT_ID = "fakeAccountID"
CLIENT_ID = "fakeClientID"
CLIENT_SECRET = "fakeClientSecret"

ZOOM_URI = "https://api.zoom.us/v2/"
ZOOM_AUTH_CALLBACK = "https://zoom.us/oauth/token"


class TestZoom(unittest.TestCase):
    @requests_mock.Mocker()
    def setUp(self, m):
        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        self.zoom = Zoom(ACCOUNT_ID, CLIENT_ID, CLIENT_SECRET)
        self.zoomv2 = Zoom(
            ACCOUNT_ID,
            CLIENT_ID,
            CLIENT_SECRET,
            parsons_version="v2",
        )

    @requests_mock.Mocker()
    def test_get_users(self, m):
        user_json = {
            "page_count": 1,
            "page_number": 1,
            "page_size": 30,
            "total_records": 1,
            "users": [
                {
                    "id": "C5A2nRWwTMm_hXyJb1JXMh",
                    "first_name": "Bob",
                    "last_name": "McBob",
                    "email": "bob@bob.com",
                    "type": 2,
                    "pmi": 8374523641,
                    "timezone": "America/New_York",
                    "verified": 1,
                    "dept": "",
                    "created_at": "2017-10-06T15:22:34Z",
                    "last_login_time": "2020-05-06T16:50:45Z",
                    "last_client_version": "",
                    "language": "",
                    "phone_number": "",
                    "status": "active",
                }
            ],
        }

        tbl = Table(
            [
                {
                    "id": "C5A2nRWwTMm_hXyJb1JXMh",
                    "first_name": "Bob",
                    "last_name": "McBob",
                    "email": "bob@bob.com",
                    "type": 2,
                    "pmi": 8374523641,
                    "timezone": "America/New_York",
                    "verified": 1,
                    "dept": "",
                    "created_at": "2017-10-06T15:22:34Z",
                    "last_login_time": "2020-05-06T16:50:45Z",
                    "last_client_version": "",
                    "language": "",
                    "phone_number": "",
                    "status": "active",
                }
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "users", json=user_json)
        assert_matching_tables(self.zoom.get_users(), tbl)

    @requests_mock.Mocker()
    def test_get_meeting_participants(self, m):
        participants = {
            "page_count": 1,
            "page_size": 30,
            "total_records": 4,
            "next_page_token": "",
            "participants": [
                {
                    "id": "",
                    "user_id": "16778240",
                    "name": "Barack Obama",
                    "user_email": "",
                    "join_time": "2020-04-24T21:00:26Z",
                    "leave_time": "2020-04-24T22:24:38Z",
                    "duration": 5052,
                    "attentiveness_score": "",
                },
                {
                    "id": "",
                    "user_id": "16779264",
                    "name": "",
                    "user_email": "",
                    "join_time": "2020-04-24T21:00:45Z",
                    "leave_time": "2020-04-24T22:24:38Z",
                    "duration": 5033,
                    "attentiveness_score": "",
                },
            ],
        }

        tbl = Table(
            [
                {
                    "id": "",
                    "user_id": "16778240",
                    "name": "Barack Obama",
                    "user_email": "",
                    "join_time": "2020-04-24T21:00:26Z",
                    "leave_time": "2020-04-24T22:24:38Z",
                    "duration": 5052,
                    "attentiveness_score": "",
                },
                {
                    "id": "",
                    "user_id": "16779264",
                    "name": "",
                    "user_email": "",
                    "join_time": "2020-04-24T21:00:45Z",
                    "leave_time": "2020-04-24T22:24:38Z",
                    "duration": 5033,
                    "attentiveness_score": "",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "report/meetings/123/participants", json=participants)
        assert_matching_tables(self.zoom.get_past_meeting_participants(123), tbl)

    @requests_mock.Mocker()
    def test_get_meeting_registrants(self, m):
        registrants = {
            "page_count": 1,
            "page_size": 30,
            "total_records": 4,
            "next_page_token": "",
            "registrants": [
                {
                    "id": "",
                    "user_id": "16778240",
                    "name": "Barack Obama",
                    "user_email": "",
                    "purchasing_time_frame": "Within a month",
                    "role_in_purchase_process": "Not involved",
                },
                {
                    "id": "",
                    "user_id": "16779264",
                    "name": "",
                    "user_email": "",
                    "purchasing_time_frame": "Within a month",
                    "role_in_purchase_process": "Not involved",
                },
            ],
        }

        tbl = Table(
            [
                {
                    "id": "",
                    "user_id": "16778240",
                    "name": "Barack Obama",
                    "user_email": "",
                    "purchasing_time_frame": "Within a month",
                    "role_in_purchase_process": "Not involved",
                },
                {
                    "id": "",
                    "user_id": "16779264",
                    "name": "",
                    "user_email": "",
                    "purchasing_time_frame": "Within a month",
                    "role_in_purchase_process": "Not involved",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "meetings/123/registrants", json=registrants)
        assert_matching_tables(self.zoom.get_meeting_registrants(123), tbl)

    @requests_mock.Mocker()
    def test_get_user_webinars(self, m):
        webinars = {
            "page_count": 1,
            "page_size": 30,
            "total_records": 4,
            "next_page_token": "",
            "webinars": [
                {
                    "uuid": "dsghfkhaewfds",
                    "id": "",
                    "host_id": "24654130000000",
                    "topic": "My Webinar",
                    "agenda": "Learn more about Zoom APIs",
                    "type": "5",
                    "duration": "60",
                    "start_time": "2019-09-24T22:00:00Z",
                    "timezone": "America/Los_Angeles",
                    "created_at": "2019-08-30T22:00:00Z",
                    "join_url": "https://zoom.us/0001000/awesomewebinar",
                },
                {
                    "uuid": "dhf8as7dhf",
                    "id": "",
                    "host_id": "24654130000345",
                    "topic": "My Webinar",
                    "agenda": "Learn more about Zoom APIs",
                    "type": "5",
                    "duration": "60",
                    "start_time": "2019-09-24T22:00:00Z",
                    "timezone": "America/Los_Angeles",
                    "created_at": "2019-08-30T22:00:00Z",
                    "join_url": "https://zoom.us/0001000/awesomewebinar",
                },
            ],
        }

        tbl = Table(
            [
                {
                    "uuid": "dsghfkhaewfds",
                    "id": "",
                    "host_id": "24654130000000",
                    "topic": "My Webinar",
                    "agenda": "Learn more about Zoom APIs",
                    "type": "5",
                    "duration": "60",
                    "start_time": "2019-09-24T22:00:00Z",
                    "timezone": "America/Los_Angeles",
                    "created_at": "2019-08-30T22:00:00Z",
                    "join_url": "https://zoom.us/0001000/awesomewebinar",
                },
                {
                    "uuid": "dhf8as7dhf",
                    "id": "",
                    "host_id": "24654130000345",
                    "topic": "My Webinar",
                    "agenda": "Learn more about Zoom APIs",
                    "type": "5",
                    "duration": "60",
                    "start_time": "2019-09-24T22:00:00Z",
                    "timezone": "America/Los_Angeles",
                    "created_at": "2019-08-30T22:00:00Z",
                    "join_url": "https://zoom.us/0001000/awesomewebinar",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "users/123/webinars", json=webinars)
        assert_matching_tables(self.zoom.get_user_webinars(123), tbl)

    @requests_mock.Mocker()
    def test_get_past_webinar_participants(self, m):
        participants = {
            "page_count": 1,
            "page_size": 30,
            "total_records": 4,
            "next_page_token": "",
            "participants": [
                {
                    "id": "",
                    "user_id": "sdfjkldsf87987",
                    "name": "Barack",
                    "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                    "join_time": "2019-02-01T12:34:12.660Z",
                    "leave_time": "2019-03-01T12:34:12.660Z",
                    "duration": "20",
                },
                {
                    "id": "",
                    "user_id": "sdfjkldsfdfgdfg",
                    "name": "Joe",
                    "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                    "join_time": "2019-02-01T12:34:12.660Z",
                    "leave_time": "2019-03-01T12:34:12.660Z",
                    "duration": "20",
                },
            ],
        }

        tbl = Table(
            [
                {
                    "id": "",
                    "user_id": "sdfjkldsf87987",
                    "name": "Barack",
                    "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                    "join_time": "2019-02-01T12:34:12.660Z",
                    "leave_time": "2019-03-01T12:34:12.660Z",
                    "duration": "20",
                },
                {
                    "id": "",
                    "user_id": "sdfjkldsfdfgdfg",
                    "name": "Joe",
                    "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                    "join_time": "2019-02-01T12:34:12.660Z",
                    "leave_time": "2019-03-01T12:34:12.660Z",
                    "duration": "20",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "report/webinars/123/participants", json=participants)
        assert_matching_tables(self.zoom.get_past_webinar_participants(123), tbl)

    @requests_mock.Mocker()
    def test_get_past_webinar_report(self, m):
        report = {
            "custom_keys": [{"key": "Host Nation", "value": "US"}],
            "dept": "HR",
            "duration": 2,
            "end_time": "2022-03-15T07:42:22Z",
            "id": 345678902224,
            "participants_count": 4,
            "start_time": "2022-03-15T07:40:46Z",
            "topic": "My Meeting",
            "total_minutes": 3,
            "tracking_fields": [{"field": "Host Nation", "value": "US"}],
            "type": 4,
            "user_email": "jchill@example.com",
            "user_name": "Jill Chill",
            "uuid": "4444AAAiAAAAAiAiAiiAii==",
        }

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "report/webinars/123", json=report)
        assert_matching_tables(self.zoom.get_past_webinar_report(123), report)

    @requests_mock.Mocker()
    def test_get_webinar_registrants(self, m):
        registrants = {
            "page_count": 1,
            "page_size": 30,
            "total_records": 4,
            "next_page_token": "",
            "registrants": [
                {
                    "id": "",
                    "email": "barack@obama.com",
                    "first_name": "Barack",
                    "last_name": "Obama",
                    "address": "dsfhkdjsfh st",
                    "city": "jackson heights",
                    "country": "US",
                    "zip": "11371",
                    "state": "NY",
                    "phone": "00000000",
                    "industry": "Food",
                    "org": "Cooking Org",
                    "job_title": "Chef",
                    "purchasing_time_frame": "1-3 months",
                    "role_in_purchase_process": "Influencer",
                    "no_of_employees": "10",
                    "comments": "Looking forward to the Webinar",
                    "custom_questions": [
                        {
                            "title": "What do you hope to learn from this Webinar?",
                            "value": "Look forward to learning how you come up with recipes and services",
                        }
                    ],
                    "status": "approved",
                    "create_time": "2019-02-26T23:01:16.899Z",
                    "join_url": "https://zoom.us/webinar/mywebinariscool",
                },
                {
                    "id": "",
                    "email": "joe@biden.com",
                    "first_name": "Joe",
                    "last_name": "Biden",
                    "address": "dsfhkdjsfh st",
                    "city": "jackson heights",
                    "country": "US",
                    "zip": "11371",
                    "state": "NY",
                    "phone": "00000000",
                    "industry": "Food",
                    "org": "Cooking Org",
                    "job_title": "Chef",
                    "purchasing_time_frame": "1-3 months",
                    "role_in_purchase_process": "Influencer",
                    "no_of_employees": "10",
                    "comments": "Looking forward to the Webinar",
                    "custom_questions": [
                        {
                            "title": "What do you hope to learn from this Webinar?",
                            "value": "Look forward to learning how you come up with recipes and services",
                        }
                    ],
                    "status": "approved",
                    "create_time": "2019-02-26T23:01:16.899Z",
                    "join_url": "https://zoom.us/webinar/mywebinariscool",
                },
            ],
        }

        tbl = Table(
            [
                {
                    "id": "",
                    "email": "barack@obama.com",
                    "first_name": "Barack",
                    "last_name": "Obama",
                    "address": "dsfhkdjsfh st",
                    "city": "jackson heights",
                    "country": "US",
                    "zip": "11371",
                    "state": "NY",
                    "phone": "00000000",
                    "industry": "Food",
                    "org": "Cooking Org",
                    "job_title": "Chef",
                    "purchasing_time_frame": "1-3 months",
                    "role_in_purchase_process": "Influencer",
                    "no_of_employees": "10",
                    "comments": "Looking forward to the Webinar",
                    "custom_questions": [
                        {
                            "title": "What do you hope to learn from this Webinar?",
                            "value": "Look forward to learning how you come up with recipes and services",
                        }
                    ],
                    "status": "approved",
                    "create_time": "2019-02-26T23:01:16.899Z",
                    "join_url": "https://zoom.us/webinar/mywebinariscool",
                },
                {
                    "id": "",
                    "email": "joe@biden.com",
                    "first_name": "Joe",
                    "last_name": "Biden",
                    "address": "dsfhkdjsfh st",
                    "city": "jackson heights",
                    "country": "US",
                    "zip": "11371",
                    "state": "NY",
                    "phone": "00000000",
                    "industry": "Food",
                    "org": "Cooking Org",
                    "job_title": "Chef",
                    "purchasing_time_frame": "1-3 months",
                    "role_in_purchase_process": "Influencer",
                    "no_of_employees": "10",
                    "comments": "Looking forward to the Webinar",
                    "custom_questions": [
                        {
                            "title": "What do you hope to learn from this Webinar?",
                            "value": "Look forward to learning how you come up with recipes and services",
                        }
                    ],
                    "status": "approved",
                    "create_time": "2019-02-26T23:01:16.899Z",
                    "join_url": "https://zoom.us/webinar/mywebinariscool",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "webinars/123/registrants", json=registrants)
        assert_matching_tables(self.zoom.get_webinar_registrants(123), tbl)

    @requests_mock.Mocker()
    def test_get_meeting_poll_metadata(self, m):
        poll = {
            "id": 1,
            "status": "started",
            "anonymous": "False",
            "poll_type": 1,
            "questions": [
                {
                    "answer_max_character": 1,
                    "answer_min_character": 1,
                    "answer_required": "False",
                    "answers": "Larry David's Curb Your Enthusiasm",
                    "case_sensitive": "True",
                    "name": "Secret Truth",
                    "prompts": [
                        {
                            "prompt_question": "What's the secret truth of the universe?",
                            "prompt_right_answers": [
                                "Pizza delivery",
                                "Larry David's Curb Your Enthusiasm",
                            ],
                        }
                    ],
                    "rating_max_label": "",
                    "rating_max_value": 1,
                    "rating_min_label": "",
                    "rating_min_value": 0,
                    "right_answers": "",
                    "show_as_dropdown": False,
                    "type": "short_answer",
                }
            ],
        }

        tbl = Table(
            [
                {
                    "answer_max_character": 1,
                    "answer_min_character": 1,
                    "answer_required": "False",
                    "answers": "Larry David's Curb Your Enthusiasm",
                    "case_sensitive": "True",
                    "name": "Secret Truth",
                    "rating_max_label": "",
                    "rating_max_value": 1,
                    "rating_min_label": "",
                    "rating_min_value": 0,
                    "right_answers": "",
                    "show_as_dropdown": False,
                    "type": "short_answer",
                    "prompts__prompt_question": "What's the secret truth of the universe?",
                    "prompts__prompt_right_answers": [
                        "Pizza delivery",
                        "Larry David's Curb Your Enthusiasm",
                    ],
                }
            ],
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "meetings/123/polls/1", json=poll)
        assert_matching_tables(self.zoom.get_meeting_poll_metadata(123, 1), tbl)

    @requests_mock.Mocker()
    def test_get_meeting_all_polls_metadata(self, m):
        polls = {
            "polls": [
                {
                    "id": 1,
                    "status": "started",
                    "anonymous": "False",
                    "poll_type": 1,
                    "questions": [
                        {
                            "answer_max_character": 1,
                            "answer_min_character": 1,
                            "answer_required": "False",
                            "answers": "Larry David's Curb Your Enthusiasm",
                            "case_sensitive": "True",
                            "name": "Secret Truth",
                            "prompts": [
                                {
                                    "prompt_question": "What's the secret truth of the universe?",
                                    "prompt_right_answers": [
                                        "Pizza delivery",
                                        "Larry David's Curb Your Enthusiasm",
                                    ],
                                }
                            ],
                            "rating_max_label": "",
                            "rating_max_value": 1,
                            "rating_min_label": "",
                            "rating_min_value": 0,
                            "right_answers": "",
                            "show_as_dropdown": False,
                            "type": "short_answer",
                        }
                    ],
                },
                {
                    "id": 2,
                    "status": "started",
                    "anonymous": "False",
                    "poll_type": 1,
                    "questions": [
                        {
                            "answer_max_character": 1,
                            "answer_min_character": 1,
                            "answer_required": "False",
                            "answers": "Mets",
                            "case_sensitive": "True",
                            "name": "Play ball",
                            "prompts": [
                                {
                                    "prompt_question": "Best NY baseball team?",
                                    "prompt_right_answers": ["Mets"],
                                }
                            ],
                            "rating_max_label": "",
                            "rating_max_value": 1,
                            "rating_min_label": "",
                            "rating_min_value": 0,
                            "right_answers": "",
                            "show_as_dropdown": True,
                            "type": "short_answer",
                        }
                    ],
                },
            ]
        }

        tbl = Table(
            [
                {
                    "id": 1,
                    "status": "started",
                    "anonymous": "False",
                    "poll_type": 1,
                    "questions__answer_max_character": 1,
                    "questions__answer_min_character": 1,
                    "questions__answer_required": "False",
                    "questions__answers": "Larry David's Curb Your Enthusiasm",
                    "questions__case_sensitive": "True",
                    "questions__name": "Secret Truth",
                    "questions__prompts": [
                        {
                            "prompt_question": "What's the secret truth of the universe?",
                            "prompt_right_answers": [
                                "Pizza delivery",
                                "Larry David's Curb Your Enthusiasm",
                            ],
                        }
                    ],
                    "questions__rating_max_label": "",
                    "questions__rating_max_value": 1,
                    "questions__rating_min_label": "",
                    "questions__rating_min_value": 0,
                    "questions__right_answers": "",
                    "questions__show_as_dropdown": False,
                    "questions__type": "short_answer",
                },
                {
                    "id": 2,
                    "status": "started",
                    "anonymous": "False",
                    "poll_type": 1,
                    "questions__answer_max_character": 1,
                    "questions__answer_min_character": 1,
                    "questions__answer_required": "False",
                    "questions__answers": "Mets",
                    "questions__case_sensitive": "True",
                    "questions__name": "Play ball",
                    "questions__prompts": [
                        {
                            "prompt_question": "Best NY baseball team?",
                            "prompt_right_answers": ["Mets"],
                        }
                    ],
                    "questions__rating_max_label": "",
                    "questions__rating_max_value": 1,
                    "questions__rating_min_label": "",
                    "questions__rating_min_value": 0,
                    "questions__right_answers": "",
                    "questions__show_as_dropdown": True,
                    "questions__type": "short_answer",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "meetings/123/polls", json=polls)
        assert_matching_tables(self.zoom.get_meeting_all_polls_metadata(123), tbl)

    @requests_mock.Mocker()
    def test_get_past_meeting_poll_metadata(self, m):
        poll = {
            "id": 1,
            "status": "started",
            "anonymous": "False",
            "poll_type": 1,
            "questions": [
                {
                    "answer_max_character": 1,
                    "answer_min_character": 1,
                    "answer_required": "False",
                    "answers": "Larry David's Curb Your Enthusiasm",
                    "case_sensitive": "True",
                    "name": "Secret Truth",
                    "question_details": [
                        {
                            "prompt_question": "What's the secret truth of the universe?",
                            "prompt_right_answers": [
                                "Pizza delivery",
                                "Larry David's Curb Your Enthusiasm",
                            ],
                        }
                    ],
                    "rating_max_label": "",
                    "rating_max_value": 1,
                    "rating_min_label": "",
                    "rating_min_value": 0,
                    "right_answers": "",
                    "show_as_dropdown": False,
                    "type": "short_answer",
                }
            ],
        }

        tbl = Table(
            [
                {
                    "answer_max_character": 1,
                    "answer_min_character": 1,
                    "answer_required": "False",
                    "answers": "Larry David's Curb Your Enthusiasm",
                    "case_sensitive": "True",
                    "name": "Secret Truth",
                    "rating_max_label": "",
                    "rating_max_value": 1,
                    "rating_min_label": "",
                    "rating_min_value": 0,
                    "right_answers": "",
                    "show_as_dropdown": False,
                    "type": "short_answer",
                    "question_details__prompt_question": "What's the secret truth of the universe?",
                    "question_details__prompt_right_answers": [
                        "Pizza delivery",
                        "Larry David's Curb Your Enthusiasm",
                    ],
                }
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "past_meetings/123/polls", json=poll)
        assert_matching_tables(self.zoom.get_past_meeting_poll_metadata(123), tbl)

    @requests_mock.Mocker()
    def test_get_webinar_poll_metadata(self, m):
        poll = {
            "id": "QalIoKWLTJehBJ8e1xRrbQ",
            "status": "notstart",
            "anonymous": True,
            "poll_type": 2,
            "questions": [
                {
                    "answer_max_character": 200,
                    "answer_min_character": 1,
                    "answer_required": False,
                    "answers": ["Extremely useful"],
                    "case_sensitive": False,
                    "name": "How useful was this meeting?",
                    "prompts": [
                        {
                            "prompt_question": "How are you?",
                            "prompt_right_answers": ["Good"],
                        }
                    ],
                    "rating_max_label": "Extremely Likely",
                    "rating_max_value": 4,
                    "rating_min_label": "Not likely",
                    "rating_min_value": 0,
                    "right_answers": ["Good"],
                    "show_as_dropdown": False,
                    "type": "single",
                }
            ],
            "title": "Learn something new",
        }

        tbl = Table(
            [
                {
                    "answer_max_character": 200,
                    "answer_min_character": 1,
                    "answer_required": False,
                    "answers": ["Extremely useful"],
                    "case_sensitive": False,
                    "name": "How useful was this meeting?",
                    "rating_max_label": "Extremely Likely",
                    "rating_max_value": 4,
                    "rating_min_label": "Not likely",
                    "rating_min_value": 0,
                    "right_answers": ["Good"],
                    "show_as_dropdown": False,
                    "type": "single",
                    "prompts__prompt_question": "How are you?",
                    "prompts__prompt_right_answers": ["Good"],
                }
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "webinars/123/polls/456", json=poll)
        assert_matching_tables(self.zoom.get_webinar_poll_metadata(123, 456), tbl)

    @requests_mock.Mocker()
    def test_get_webinar_all_polls_metadata(self, m):
        polls = {
            "polls": [
                {
                    "id": "QalIoKWLTJehBJ8e1xRrbQ",
                    "status": "notstart",
                    "anonymous": True,
                    "poll_type": 2,
                    "questions": [
                        {
                            "answer_max_character": 200,
                            "answer_min_character": 1,
                            "answer_required": False,
                            "answers": ["Extremely useful"],
                            "case_sensitive": False,
                            "name": "How useful was this meeting?",
                            "prompts": [
                                {
                                    "prompt_question": "How are you?",
                                    "prompt_right_answers": ["Good"],
                                }
                            ],
                            "rating_max_label": "Extremely Likely",
                            "rating_max_value": 4,
                            "rating_min_label": "Not likely",
                            "rating_min_value": 0,
                            "right_answers": ["Good"],
                            "show_as_dropdown": False,
                            "type": "single",
                        }
                    ],
                    "title": "Learn something new",
                }
            ],
            "total_records": 1,
        }

        tbl = Table(
            [
                {
                    "id": "QalIoKWLTJehBJ8e1xRrbQ",
                    "status": "notstart",
                    "anonymous": True,
                    "poll_type": 2,
                    "title": "Learn something new",
                    "questions__answer_max_character": 200,
                    "questions__answer_min_character": 1,
                    "questions__answer_required": False,
                    "questions__answers": ["Extremely useful"],
                    "questions__case_sensitive": False,
                    "questions__name": "How useful was this meeting?",
                    "questions__prompts": [
                        {
                            "prompt_question": "How are you?",
                            "prompt_right_answers": ["Good"],
                        }
                    ],
                    "questions__rating_max_label": "Extremely Likely",
                    "questions__rating_max_value": 4,
                    "questions__rating_min_label": "Not likely",
                    "questions__rating_min_value": 0,
                    "questions__right_answers": ["Good"],
                    "questions__show_as_dropdown": False,
                    "questions__type": "single",
                }
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "webinars/123/polls", json=polls)
        assert_matching_tables(self.zoom.get_webinar_all_polls_metadata(123), tbl)

    @requests_mock.Mocker()
    def test_get_past_webinar_poll_metadata(self, m):
        poll = {
            "id": "QalIoKWLTJehBJ8e1xRrbQ",
            "status": "notstart",
            "anonymous": True,
            "poll_type": 2,
            "questions": [
                {
                    "answer_max_character": 200,
                    "answer_min_character": 1,
                    "answer_required": False,
                    "answers": ["Extremely useful"],
                    "case_sensitive": False,
                    "name": "How useful was this meeting?",
                    "question_details": [
                        {
                            "question": "How are you?",
                            "answer": ["Good"],
                        }
                    ],
                    "rating_max_label": "Extremely Likely",
                    "rating_max_value": 4,
                    "rating_min_label": "Not likely",
                    "rating_min_value": 0,
                    "right_answers": ["Good"],
                    "show_as_dropdown": False,
                    "type": "single",
                }
            ],
            "title": "Learn something new",
        }

        tbl = Table(
            [
                {
                    "answer_max_character": 200,
                    "answer_min_character": 1,
                    "answer_required": False,
                    "answers": ["Extremely useful"],
                    "case_sensitive": False,
                    "name": "How useful was this meeting?",
                    "rating_max_label": "Extremely Likely",
                    "rating_max_value": 4,
                    "rating_min_label": "Not likely",
                    "rating_min_value": 0,
                    "right_answers": ["Good"],
                    "show_as_dropdown": False,
                    "type": "single",
                    "question_details__answer": ["Good"],
                    "question_details__question": "How are you?",
                }
            ],
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "past_webinars/123/polls", json=poll)
        assert_matching_tables(self.zoom.get_past_webinar_poll_metadata(123), tbl)

    @requests_mock.Mocker()
    def test_get_meeting_poll_results(self, m):
        poll = {
            "id": 123456,
            "uuid": "4444AAAiAAAAAiAiAiiAii==",
            "start_time": "2022-02-01T12:34:12.660Z",
            "questions": [
                {
                    "email": "jchill@example.com",
                    "name": "Jill Chill",
                    "first_name": "Jill",
                    "last_name": "Chill",
                    "question_details": [
                        {
                            "answer": "I am wonderful.",
                            "date_time": "2022-02-01T12:37:12.660Z",
                            "polling_id": "798fGJEWrA",
                            "question": "How are you?",
                        }
                    ],
                }
            ],
        }

        tbl = Table(
            [
                {
                    "email": "jchill@example.com",
                    "name": "Jill Chill",
                    "first_name": "Jill",
                    "last_name": "Chill",
                    "answer": "I am wonderful.",
                    "date_time": "2022-02-01T12:37:12.660Z",
                    "polling_id": "798fGJEWrA",
                    "question": "How are you?",
                }
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "report/meetings/123/polls", json=poll)
        assert_matching_tables(self.zoom.get_meeting_poll_results(123), tbl)

    @requests_mock.Mocker()
    def test_get_webinar_poll_results(self, m):
        poll = {
            "id": 123456,
            "questions": [
                {
                    "email": "jchill@example.com",
                    "name": "Jill Chill",
                    "first_name": "Jill",
                    "last_name": "Chill",
                    "question_details": [
                        {
                            "answer": "I am wonderful.",
                            "date_time": "2022-02-01T12:37:12.660Z",
                            "polling_id": "798fGJEWrA",
                            "question": "How are you?",
                        }
                    ],
                }
            ],
            "start_time": "2022-02-01T12:34:12.660Z",
            "uuid": "4444AAAiAAAAAiAiAiiAii==",
        }

        tbl = Table(
            [
                {
                    "email": "jchill@example.com",
                    "name": "Jill Chill",
                    "first_name": "Jill",
                    "last_name": "Chill",
                    "answer": "I am wonderful.",
                    "date_time": "2022-02-01T12:37:12.660Z",
                    "polling_id": "798fGJEWrA",
                    "question": "How are you?",
                }
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "report/webinars/123/polls", json=poll)
        assert_matching_tables(self.zoom.get_webinar_poll_results(123), tbl)

    @requests_mock.Mocker()
    def test_get_webinar_occurrences(self, m):
        occurrences = {
            "occurrences": [
                {
                    "occurrence_id": "123456789",
                    "start_time": "2023-01-01T10:00:00Z",
                    "duration": 60,
                    "status": "available",
                },
                {
                    "occurrence_id": "987654321",
                    "start_time": "2023-01-02T14:00:00Z",
                    "duration": 90,
                    "status": "available",
                },
            ]
        }

        tbl = Table(
            [
                {
                    "occurrence_id": "123456789",
                    "start_time": "2023-01-01T10:00:00Z",
                    "duration": 60,
                    "status": "available",
                },
                {
                    "occurrence_id": "987654321",
                    "start_time": "2023-01-02T14:00:00Z",
                    "duration": 90,
                    "status": "available",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "webinars/123/", json=occurrences)
        assert_matching_tables(self.zoomv2.get_webinar_occurrences(123), tbl)



"""NGPVAN Contact Notes Endpoints"""

import logging

from parsons.etl.table import Table

logger = logging.getLogger(__name__)


class ContactNotes(object):
    def __init__(self, van_connection):
        self.connection = van_connection

    def get_contact_notes(self, van_id):
        """
        Get custom fields.

        `Args:`
            van_id : str
                VAN ID for the person to get notes for.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request(f"people/{van_id}/notes"))
        logger.info(f"Found {tbl.num_rows} custom fields.")
        return tbl

    def create_contact_note(
        self,
        van_id,
        text,
        is_view_restricted,
        note_category_id=None,
        contact_type_id=None,
        input_type_id=None,
        date_canvassed=None,
        result_code_id=None,
    ):
        """
        Create a contact note

        `Args:`
            van_id: str
                VAN ID for the person this note will be applied to.
            text: str
                The content of the note.
            is_view_restricted: bool
                Set to true if the note should be restricted only to certain users within
                the current context; set to false if the note may be viewed by any user
                in the current context.
            note_category_id: int
                Optional; if set, the note category for this note.
            contact_type_id: str
                Defaults to 82 if no value is set. This value results in a null contact type in EA.
            input_type_id: str
                Defaults to 11 if no value is set. If the value is 11,
                the input type in EA will be listed as "API"
            date_canvassed: date
                Defaults to current date if no value is set. Dates should be formatted in ISO8601 standard.
            result_code_id: str
                Defaults to 205 if no value is set. This value results in a "Contacted" result in EA.
        `Returns:`
            int
              The note ID.
        """
        note = {"text": text, "isViewRestricted": is_view_restricted}
        if note_category_id is not None:
            note["category"] = {"noteCategoryId": note_category_id}

        contact_history = {}

        if contact_type_id is not None:
            contact_history["contact_type_id"] = str(contact_type_id)
        if input_type_id is not None:
            contact_history["input_type_id"] = str(input_type_id)
        if date_canvassed is not None:
            contact_history["dateCanvassed"] = date_canvassed
        if result_code_id is not None:
            contact_history["result_code_id"] = str(result_code_id)

        if contact_history:
            note["contactHistory"] = contact_history

        r = self.connection.post_request(f"people/{van_id}/notes", json=note)
        logger.info(f"Contact note {r} created.")
        return r



# Additional padding to add on to the maximum column width to account
# for the addition of future data sets.
VARCHAR_PAD = 0.25

COL_NAME_MAX_LEN = 64

# List of varchar lengths to use for columns -- this list needs to be in order from smallest to
# largest
VARCHAR_STEPS = [32, 64, 128, 256, 512, 1024, 4096, 8192, 16384]

# Max length of a Redshift VARCHAR column
VARCHAR_MAX = 65535

# The following values are the minimum and maximum values for MySQL int
# types. https://dev.mysql.com/doc/refman/8.0/en/integer-types.html
SMALLINT_VALUES = [-32768, 32767]
MEDIUMINT_VALUES = [-8388608, 8388607]
INT_VALUES = [-2147483648, 2147483647]



# Adapted from Gmail API tutorial https://developers.google.com/gmail/api
import io
import logging
import mimetypes
import os
from abc import ABC, abstractmethod
from email.encoders import encode_base64
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import parseaddr

from validate_email import validate_email

# BUG: can't send files equal to or larger than 6MB
# There is a possible fix
# `BUG-1` are the changes that would need to be made

# BUG-1
# import httplib2shim
# import sys

logger = logging.getLogger(__name__)


class SendMail(ABC):
    """SendMail base class for sending emails.

    This class is not designed to be used directly,
    as it has useful methods for composing messages and validating emails
    but does not contain all the required functionality in order
    to send a message. Rather it should be subclassed for each different type of
    email service, and those subclasses should define an __init__
    method (to set any instance attributes such as credentials) and a _send_message
    method (to implement the actual sending of the message).

    For an example of this subclassing in practice, look at the Gmail notification
    connector in parsons.notifications.gmail.
    """

    log = logger

    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def _send_message(self, message):
        pass

    def _create_message_simple(self, sender, to, subject, message_text):
        """Create a text-only message for an email.

        `Args:`
            sender: str
                Email address of the sender.
            to: str
                Email address(es) of the recipient(s).
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
        `Returns:`
            An object passable to send_message to send
        """
        self.log.info("Creating a simple message...")

        message = MIMEText(message_text)
        message["to"] = to
        message["from"] = sender
        message["subject"] = subject

        return message

    def _create_message_html(self, sender, to, subject, message_text, message_html):
        """Create an html message for an email.

        `Args:`
            sender: str
                Email address of the sender.
            to: str
                Email address(es) of the recipient(s).
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
            message_html: str
                The html formatted text of the email message.
        `Returns:`
            An object passable to send_message to send
        """
        self.log.info("Creating an html message...")

        message = MIMEMultipart("alternative")
        message["subject"] = subject
        message["from"] = sender
        message["to"] = to
        if message_text:
            message.attach(MIMEText(message_text, "plain"))
        message.attach(MIMEText(message_html, "html"))

        return message

    def _create_message_attachments(
        self, sender, to, subject, message_text, files, message_html=None
    ):
        """Create a message for an email that includes an attachment.

        `Args:`
            sender: str
                Email address of the sender.
            to: str
                Email address of the receiver.
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
            files: list
                The path(s) to the file(s) to be attached.
            message_html: str
                Optional; The html formatted text of the email message.
        `Returns:`
            An object passable to send_message to send
        """
        self.log.info("Creating a message with attachments...")

        message = MIMEMultipart("alternative")
        message["to"] = to
        message["from"] = sender
        message["subject"] = subject

        msg = MIMEText(message_text, "plain")
        message.attach(msg)

        if message_html:
            html = MIMEText(message_html, "html")
            message.attach(html)

        for f in files:
            filename = getattr(f, "name", "file")
            file_bytes = b""

            if isinstance(f, io.StringIO):
                file_bytes = f.getvalue().encode()
            elif isinstance(f, io.BytesIO):
                file_bytes = f.getvalue()
            else:
                filename = os.path.basename(f)
                fp = open(f, "rb")
                file_bytes = fp.read()
                fp.close()

            content_type, encoding = mimetypes.guess_type(filename)
            self.log.debug(f"(File: {f}, Content-type: {content_type}, Encoding: {encoding})")

            if content_type is None or encoding is not None:
                content_type = "application/octet-stream"

            main_type, sub_type = content_type.split("/", 1)

            if main_type == "text":
                self.log.info("Added a text file.")
                msg = MIMEText(file_bytes, _subtype=sub_type, _charset="utf-8")

            elif main_type == "image":
                self.log.info("Added an image file.")
                msg = MIMEImage(file_bytes, _subtype=sub_type)
                msg.add_header("Content-ID", f"<{filename}>")

            elif main_type == "audio":
                self.log.info("Added an audio file.")
                msg = MIMEAudio(file_bytes, _subtype=sub_type)

            elif main_type == "application":
                self.log.info("Added an application file.")
                msg = MIMEApplication(file_bytes, _subtype=sub_type)

            else:
                self.log.info("Added an unknown-type file.")
                msg = MIMEBase(main_type, sub_type)
                msg.set_payload(file_bytes)
                encode_base64(msg)

            msg.add_header("Content-Disposition", "attachment", filename=filename)
            message.attach(msg)

        return message

    def _validate_email_string(self, str):
        self.log.debug(f"Validating email {str}...")
        realname, email_addr = parseaddr(str)

        if not email_addr:
            raise ValueError("Invalid email address.")

        if not validate_email(email_addr):
            raise ValueError("Invalid email address.")

        return True

    def send_email(self, sender, to, subject, message_text, message_html=None, files=None):
        """Send an email message.

        `Args:`
            sender: str
                Email address of the sender.
            to: str or list
                Email address(es) of the receiver(s). Must be in correct email
                string syntax. For example, `name@email.com` or
                `"Name" <email@email.com>`.
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
            message_html: str
                The html formatted text of the email message. If ommitted, the
                email is sent a text-only body.
            files: str or list
                The path to the file(s) to be attached.

        `Returns:`
            None
        """
        self.log.info("Preparing to send an email...")

        self.log.info("Validating email(s)")
        if isinstance(to, list):
            if len(to) == 0:
                raise EmptyListError("Must contain at least 1 email.")

            for e in to:
                self._validate_email_string(e)

            to = ", ".join(to)

        elif isinstance(to, str):
            self._validate_email_string(to)

        if not message_html and not files:
            msg_type = "simple"
            msg = self._create_message_simple(sender, to, subject, message_text)

        elif not files:
            msg_type = "html"
            msg = self._create_message_html(sender, to, subject, message_text, message_html)
        else:
            msg_type = "attachments"
            if isinstance(files, str):
                files = [files]

            msg = self._create_message_attachments(
                sender, to, subject, message_text, files, message_html
            )

        self.log.info(f"Sending a(n) {msg_type} email...")

        self._send_message(msg)

        self.log.info("Email sent succesfully.")


class EmptyListError(IndexError):
    """Throw when a list is empty that should contain at least 1 element."""

    pass



import os
from typing import List, Optional, Type, Union

from parsons.databases.database_connector import DatabaseConnector
from parsons.databases.mysql import MySQL
from parsons.databases.postgres import Postgres
from parsons.databases.redshift import Redshift
from parsons.google.google_bigquery import GoogleBigQuery as BigQuery


def discover_database(
    default_connector: Optional[
        Union[Type[DatabaseConnector], List[Type[DatabaseConnector]]]
    ] = None,
) -> DatabaseConnector:
    """Create an appropriate ``DatabaseConnector`` based on environmental variables.

    Will search the environmental variables for the proper credentials for the
    Redshift, MySQL, Postgres, and BigQuery connectors. See the documentation
    for the connectors to variables required to initialize them.

    If no suitable configuration is found, will raise an error.

    If multiple suitable configurations are found, will raise an error unless
    a default connector class or list of classes is provided.

    Note that the variables to be searched for are hard-coded in this function,
    since they are unlikely to change. If that is done, for some reason, or a
    new database connector is added, ``discover_database`` should be updated

    Args:
        default_connector: Optional, single Class or list of Classes inheriting from
        DatabaseConnector to be used as default in case multiple database configurations
        are detected.

    Returns:
        DatabaseConnector: The database connector configured in the environment.
    """
    connectors = {
        "Redshift": Redshift,
        "MySQL": MySQL,
        "Postgres": Postgres,
        "GoogleBigQuery": BigQuery,
    }

    password_vars = {
        "Redshift": "REDSHIFT_PASSWORD",
        "MySQL": "MYSQL_PASSWORD",
        "Postgres": "PGPASSWORD",
        "GoogleBigQuery": "GOOGLE_APPLICATION_CREDENTIALS",
    }

    detected = [name for name in connectors.keys() if os.getenv(password_vars[name])]

    if len(detected) > 1:
        if default_connector is None:
            raise EnvironmentError(
                f"Multiple database configurations detected: {detected}."
                " Please specify a default connector."
            )

        if isinstance(default_connector, list):
            for connector in default_connector:
                if connector.__name__ in detected:
                    return connector()
            raise EnvironmentError(
                f"None of the default connectors {default_connector} were detected."
            )
        elif default_connector.__name__ in detected:
            return default_connector()
        else:
            raise EnvironmentError(
                f"Default connector {default_connector} not detected. Detected: {detected}."
            )

    elif detected:
        return connectors[detected[0]]()
    else:
        raise EnvironmentError("Could not find any database configuration.")



import copy
import os
import unittest

import requests_mock

from parsons import CapitolCanary
from test.utils import validate_list

adv_json = {
    "data": [
        {
            "id": 7125439,
            "prefix": "Mr.",
            "firstname": "Bob",
            "middlename": "Smith",
            "lastname": "Smit",
            "suffix": None,
            "notes": None,
            "stage": None,
            "connections": 3,
            "tags": [
                "register-to-vote-38511",
                "registered-to-vote-for-2018-38511",
            ],
            "created_at": "2017-05-23 23:36:04.000000",
            "updated_at": "2018-12-17 21:55:24.000000",
            "address": {
                "street1": "25255 Maine Ave",
                "street2": "",
                "city": "Los Angeles",
                "state": "CA",
                "zip5": 96055,
                "zip4": 9534,
                "county": "Tehama",
                "latitude": "50.0632635",
                "longitude": "-122.09654",
            },
            "districts": {
                "congressional": "1",
                "stateSenate": "4",
                "stateHouse": "3",
                "cityCouncil": None,
            },
            "ids": [],
            "memberships": [
                {
                    "id": 15151443,
                    "campaignid": 25373,
                    "name": "20171121 Businesses for Responsible Tax Reform - Contact Congress",
                    "source": None,
                    "created_at": "2017-11-21 23:28:30.000000",
                },
                {
                    "id": 20025582,
                    "campaignid": 32641,
                    "name": "20180524 March for America",
                    "source": None,
                    "created_at": "2018-05-24 21:09:49.000000",
                },
            ],
            "fields": [],
            "phones": [{"id": 10537860, "address": "+19995206447", "subscribed": "false"}],
            "emails": [
                {"id": 10537871, "address": "N@k.com", "subscribed": "false"},
                {"id": 10950446, "address": "email@me.com", "subscribed": "false"},
            ],
        }
    ],
    "pagination": {
        "count": 1,
        "per_page": 100,
        "current_page": 1,
        "next_url": "https://api.phone2action.com/2.0/advocates?page=2",
    },
}

camp_json = [
    {
        "id": 25373,
        "name": "20171121 Businesses for Responsible Tax Reform - Contact Congress",
        "display_name": "Businesses for Responsible Tax Reform",
        "subtitle": "Tell Congress: Stand up for responsible tax reform!",
        "public": 1,
        "topic": None,
        "type": "CAMPAIGN",
        "link": "http://p2a.co/KHcUyTK",
        "restrict_allow": None,
        "content": {
            "summary": "",
            "introduction": "Welcome",
            "call_to_action": "Contact your officials in one click!",
            "thank_you": "<p>Thanks for taking action. Please encourage others to act by "
            "sharing on social media.</p>",
            "background_image": None,
        },
        "updated_at": {
            "date": "2017-11-21 23:27:11.000000",
            "timezone_type": 3,
            "timezone": "UTC",
        },
    }
]


def parse_request_body(m):
    kvs = m.split("&")
    return {kv.split("=")[0]: kv.split("=")[1] for kv in kvs}


class TestP2A(unittest.TestCase):
    def setUp(self):
        self.cc = CapitolCanary(app_id="an_id", app_key="app_key")

    def tearDown(self):
        pass

    def test_init_args(self):
        # Test initializing class with args
        # Done in the setUp

        pass

    def test_old_init_envs(self):
        # Test initializing class with old envs

        os.environ["PHONE2ACTION_APP_ID"] = "id"
        os.environ["PHONE2ACTION_APP_KEY"] = "key"

        cc_envs = CapitolCanary()
        self.assertEqual(cc_envs.app_id, "id")
        self.assertEqual(cc_envs.app_key, "key")

    def test_init_envs(self):
        # Test initializing class with envs

        os.environ["CAPITOLCANARY_APP_ID"] = "id"
        os.environ["CAPITOLCANARY_APP_KEY"] = "key"

        cc_envs = CapitolCanary()
        self.assertEqual(cc_envs.app_id, "id")
        self.assertEqual(cc_envs.app_key, "key")

    @requests_mock.Mocker()
    def test_get_advocates(self, m):
        m.get(self.cc.client.uri + "advocates", json=adv_json)

        adv_exp = [
            "id",
            "prefix",
            "firstname",
            "middlename",
            "lastname",
            "suffix",
            "notes",
            "stage",
            "connections",
            "created_at",
            "updated_at",
            "address_city",
            "address_county",
            "address_latitude",
            "address_longitude",
            "address_state",
            "address_street1",
            "address_street2",
            "address_zip4",
            "address_zip5",
            "districts_cityCouncil",
            "districts_congressional",
            "districts_stateHouse",
            "districts_stateSenate",
        ]

        self.assertTrue(validate_list(adv_exp, self.cc.get_advocates()["advocates"]))
        ids_exp = ["advocate_id", "ids"]

        self.assertTrue(validate_list(ids_exp, self.cc.get_advocates()["ids"]))

        phone_exp = ["advocate_id", "phones_address", "phones_id", "phones_subscribed"]
        self.assertTrue(validate_list(phone_exp, self.cc.get_advocates()["phones"]))

        tags_exp = ["advocate_id", "tags"]
        self.assertTrue(validate_list(tags_exp, self.cc.get_advocates()["tags"]))

        email_exp = ["advocate_id", "emails_address", "emails_id", "emails_subscribed"]
        self.assertTrue(validate_list(email_exp, self.cc.get_advocates()["emails"]))

        member_exp = [
            "advocate_id",
            "memberships_campaignid",
            "memberships_created_at",
            "memberships_id",
            "memberships_name",
            "memberships_source",
        ]
        self.assertTrue(validate_list(member_exp, self.cc.get_advocates()["memberships"]))

        fields_exp = ["advocate_id", "fields"]
        self.assertTrue(validate_list(fields_exp, self.cc.get_advocates()["fields"]))

    @requests_mock.Mocker()
    def test_get_advocates__by_page(self, m):
        response = copy.deepcopy(adv_json)
        # Make it look like there's more data
        response["pagination"]["count"] = 100

        m.get(self.cc.client.uri + "advocates?page=1", json=adv_json)
        m.get(
            self.cc.client.uri + "advocates?page=2",
            exc=Exception("Should only call once"),
        )

        results = self.cc.get_advocates(page=1)
        self.assertTrue(results["advocates"].num_rows, 1)

    @requests_mock.Mocker()
    def test_get_advocates__empty(self, m):
        response = copy.deepcopy(adv_json)
        response["data"] = []
        # Make it look like there's more data
        response["pagination"]["count"] = 0

        m.get(self.cc.client.uri + "advocates", json=adv_json)

        results = self.cc.get_advocates()
        self.assertTrue(results["advocates"].num_rows, 0)

    @requests_mock.Mocker()
    def test_get_campaigns(self, m):
        camp_exp = [
            "id",
            "name",
            "display_name",
            "subtitle",
            "public",
            "topic",
            "type",
            "link",
            "restrict_allow",
            "updated_at_date",
            "updated_at_timezone",
            "updated_at_timezone_type",
            "content_background_image",
            "content_call_to_action",
            "content_introduction",
            "content_summary",
            "content_thank_you",
        ]

        m.get(self.cc.client.uri + "campaigns", json=camp_json)

        self.assertTrue(validate_list(camp_exp, self.cc.get_campaigns()))

    @requests_mock.Mocker()
    def test_create_advocate(self, m):
        m.post(self.cc.client.uri + "advocates", json={"advocateid": 1})

        # Test arg validation - create requires a phone or an email
        self.assertRaises(
            ValueError,
            lambda: self.cc.create_advocate(campaigns=[1], firstname="Foo", lastname="bar"),
        )
        # Test arg validation - sms opt in requires a phone
        self.assertRaises(
            ValueError,
            lambda: self.cc.create_advocate(campaigns=[1], email="foo@bar.com", sms_optin=True),
        )

        # Test arg validation - email opt in requires a email
        self.assertRaises(
            ValueError,
            lambda: self.cc.create_advocate(campaigns=[1], phone="1234567890", email_optin=True),
        )

        # Test a successful call
        advocateid = self.cc.create_advocate(
            campaigns=[1], email="foo@bar.com", email_optin=True, firstname="Test"
        )
        self.assertTrue(m.called)
        self.assertEqual(advocateid, 1)

        # Check that the properties were mapped
        data = parse_request_body(m.last_request.text)
        self.assertEqual(data["firstname"], "Test")
        self.assertNotIn("lastname", data)
        self.assertEqual(data["emailOptin"], "1")
        self.assertEqual(data["email"], "foo%40bar.com")

    @requests_mock.Mocker()
    def test_update_advocate(self, m):
        m.post(self.cc.client.uri + "advocates")

        # Test arg validation - sms opt in requires a phone
        self.assertRaises(
            ValueError, lambda: self.cc.update_advocate(advocate_id=1, sms_optin=True)
        )

        # Test arg validation - email opt in requires a email
        self.assertRaises(
            ValueError, lambda: self.cc.update_advocate(advocate_id=1, email_optin=True)
        )

        # Test a successful call
        self.cc.update_advocate(
            advocate_id=1,
            campaigns=[1],
            email="foo@bar.com",
            email_optin=True,
            firstname="Test",
        )
        self.assertTrue(m.called)

        # Check that the properties were mapped
        data = parse_request_body(m.last_request.text)
        self.assertEqual(data["firstname"], "Test")
        self.assertNotIn("lastname", data)
        self.assertEqual(data["emailOptin"], "1")
        self.assertEqual(data["email"], "foo%40bar.com")



import logging

from parsons.etl.table import Table

logger = logging.getLogger(__name__)


class DBSync:
    """
    Sync tables between databases. Works with ``Postgres``, ``Redshift``, ``MySQL``
    databases.

    `Args:`
        source_db: Database connection object
            A database object.
        destination_db: Database connection object
            A database object.
        read_chunk_size: int
            The number of rows to read from the source at a time when syncing a table. The
            default value is 100,000 rows.
        write_chunk_size: int
            The number of rows to batch up before writing out to the destination. This value
            defaults to whatever the read_chunk_size is.
        retries: int
            The number of times to retry if there is an error processing a
            chunk of data. The default value is 0.
    `Returns:`
        A DBSync object.
    """

    def __init__(
        self,
        source_db,
        destination_db,
        read_chunk_size=100_000,
        write_chunk_size=None,
        retries=0,
    ):
        self.source_db = source_db
        self.dest_db = destination_db
        self.read_chunk_size = read_chunk_size
        self.write_chunk_size = write_chunk_size or read_chunk_size
        self.retries = retries

    def table_sync_full(
        self,
        source_table,
        destination_table,
        if_exists="drop",
        order_by=None,
        verify_row_count=True,
        **kwargs,
    ):
        """
        Full sync of table from a source database to a destination database. This will
        wipe all data from the destination table.

        `Args:`
            source_table: str
                Full table path (e.g. ``my_schema.my_table``)
            destination_table: str
                Full table path (e.g. ``my_schema.my_table``)
            if_exists: str
                If destination table exists either ``drop``, ``truncate``, or ``drop_if_needed``.
                Truncate is useful when there are dependent views associated with the table.
                Drop if needed defaults to ``truncate``, but if an error occurs (because a data
                type or length has changed), it will instead ``drop``.
            order_by: str
                Name of the column to order rows by to ensure stable sorting of results across
                chunks.
            verify_row_count: bool
                Whether or not to verify the count of rows in the source and destination table
                are the same at the end of the sync.
            **kwargs: args
                Optional copy arguments for destination database.
        `Returns:`
            ``None``
        """

        # Create the table objects
        source_tbl = self.source_db.table(source_table)
        destination_tbl = self.dest_db.table(destination_table)

        logger.info(f"Syncing full table data from {source_table} to {destination_table}")

        # Drop or truncate if the destination table exists
        if destination_tbl.exists:
            if if_exists == "drop":
                destination_tbl.drop()
            elif if_exists == "truncate":
                self._check_column_match(source_tbl, destination_tbl)
                destination_tbl.truncate()
            elif if_exists == "drop_if_needed":
                try:
                    self._check_column_match(source_tbl, destination_tbl)
                    destination_tbl.truncate()
                except Exception:
                    logger.info(f"needed to drop {destination_tbl}...")
                    destination_tbl.drop()
            else:
                raise ValueError("Invalid if_exists argument. Must be drop or truncate.")

        # Create the table, if needed.
        if not destination_tbl.exists:
            self.create_table(source_table, destination_table)

        copied_rows = self.copy_rows(source_table, destination_table, None, order_by, **kwargs)

        if verify_row_count:
            self._row_count_verify(source_tbl, destination_tbl)

        logger.info(f"{source_table} synced: {copied_rows} total rows copied.")

    def table_sync_incremental(
        self,
        source_table,
        destination_table,
        primary_key,
        distinct_check=True,
        verify_row_count=True,
        **kwargs,
    ):
        """
        Incremental sync of table from a source database to a destination database
        using an incremental primary key.

        `Args:`
            source_table: str
                Full table path (e.g. ``my_schema.my_table``)
            destination_table: str
                Full table path (e.g. ``my_schema.my_table``)
            primary_key: str
                The name of the primary key. This must be the same for the source and
                destination table.
            distinct_check: bool
                Check that the source table primary key is distinct prior to running the
                sync. If it is not, an error will be raised.
            verify_row_count: bool
                Whether or not to verify the count of rows in the source and destination table
                are the same at the end of the sync.
            **kwargs: args
                Optional copy arguments for destination database.
        `Returns:`
            ``None``
        """

        # Create the table objects
        source_tbl = self.source_db.table(source_table)
        destination_tbl = self.dest_db.table(destination_table)

        # Check that the destination table exists. If it does not, then run a
        # full sync instead.
        if not destination_tbl.exists:
            logger.info(
                "Destination tables %s does not exist, running a full sync",
                destination_table,
            )
            self.table_sync_full(
                source_table,
                destination_table,
                order_by=primary_key,
                verify_row_count=verify_row_count,
                **kwargs,
            )
            return

        # Check that the source table primary key is distinct
        if distinct_check and not source_tbl.distinct_primary_key(primary_key):
            logger.info(
                "Checking for distinct values for column %s in table %s",
                primary_key,
                source_table,
            )
            raise ValueError("{primary_key} is not distinct in source table.")

        # Get the max source table and destination table primary key
        logger.debug(
            "Calculating the maximum value for %s for source table %s",
            primary_key,
            source_table,
        )
        source_max_pk = source_tbl.max_primary_key(primary_key)
        logger.debug(
            "Calculating the maximum value for %s for destination table %s",
            primary_key,
            destination_table,
        )
        dest_max_pk = destination_tbl.max_primary_key(primary_key)

        # Check for a mismatch in row counts; if dest_max_pk is None, or destination is empty
        # and we don't have to worry about this check.
        if dest_max_pk is not None and dest_max_pk > source_max_pk:
            raise ValueError("Destination DB primary key greater than source DB primary key.")

        # Do not copied if row counts are equal.
        elif dest_max_pk == source_max_pk:
            logger.info("Tables are already in sync.")
            return None

        else:
            rows_copied = self.copy_rows(
                source_table, destination_table, dest_max_pk, primary_key, **kwargs
            )

            logger.info("Copied %s new rows to %s.", rows_copied, destination_table)

        if verify_row_count:
            self._row_count_verify(source_tbl, destination_tbl)

        logger.info(f"{source_table} synced to {destination_table}.")

    def copy_rows(self, source_table_name, destination_table_name, cutoff, order_by, **kwargs):
        """
        Copy the rows from the source to the destination.

        `Args:`
            source_table_name: str
                Full table path (e.g. ``my_schema.my_table``)
            destination_table_name: str
                Full table path (e.g. ``my_schema.my_table``)
            cutoff:
                Start value to use as a minimum for incremental updates.
            order_by:
                Column to use to order the data to ensure a stable sort.
            **kwargs: args
                Optional copy arguments for destination database.
        `Returns:`
            ``None``
        """

        # Create the table objects
        source_table = self.source_db.table(source_table_name)

        # Initialize the Parsons table we will use to store rows before writing
        buffer = Table()

        # Track the number of retries we have left before giving up
        retries_left = self.retries + 1

        total_rows_downloaded = 0
        total_rows_written = 0
        rows_buffered = 0

        # Keep going until we break out
        while True:
            try:
                # Get the records to load into the database
                if cutoff:
                    # If we have a cutoff, we are loading data incrementally -- filter out
                    # any data before our cutoff
                    rows = source_table.get_new_rows(
                        primary_key=order_by,
                        cutoff_value=cutoff,
                        offset=total_rows_downloaded,
                        chunk_size=self.read_chunk_size,
                    )
                else:
                    # Get a chunk
                    rows = source_table.get_rows(
                        offset=total_rows_downloaded,
                        chunk_size=self.read_chunk_size,
                        order_by=order_by,
                    )

                number_of_rows = rows.num_rows
                total_rows_downloaded += number_of_rows

                # If we didn't get any data, exit the loop -- there's nothing to load
                if number_of_rows == 0:
                    # If we have any rows that are unwritten, flush them to the destination database
                    if rows_buffered > 0:
                        logger.debug(
                            "Copying %s rows to %s",
                            rows_buffered,
                            destination_table_name,
                        )
                        self.dest_db.copy(
                            buffer, destination_table_name, if_exists="append", **kwargs
                        )
                        total_rows_written += rows_buffered

                        # Reset the buffer
                        rows_buffered = 0
                        buffer = Table()

                    break

                # Add the new rows to our buffer
                buffer.concat(rows)
                rows_buffered += number_of_rows

                # If our buffer reaches our write threshold, write it out
                if rows_buffered >= self.write_chunk_size:
                    logger.debug("Copying %s rows to %s", rows_buffered, destination_table_name)
                    self.dest_db.copy(buffer, destination_table_name, if_exists="append", **kwargs)
                    total_rows_written += rows_buffered

                    # Reset the buffer
                    rows_buffered = 0
                    buffer = Table()

            except Exception:
                # Tick down the number of retries
                retries_left -= 1

                # If we are out of retries, fail
                if retries_left == 0:
                    logger.debug("No retries remaining")
                    raise

                # Otherwise, log the exception and try again
                logger.exception("Unhandled error copying data; retrying")

        return total_rows_written

    @staticmethod
    def _check_column_match(source_table_obj, destination_table_obj):
        """
        Ensure that the columns from each table match
        """

        if source_table_obj.columns != destination_table_obj.columns:
            raise ValueError(
                """Destination table columns do not match source table columns.
                             Consider dropping destination table and running a full sync."""
            )

    @staticmethod
    def _row_count_verify(source_table_obj, destination_table_obj):
        """
        Ensure the the rows of the source table and the destination table match
        """

        source_row_count = source_table_obj.num_rows
        dest_row_count = destination_table_obj.num_rows

        if source_row_count != dest_row_count:
            logger.warning(
                (
                    f"Table count mismatch. Source table contains {source_row_count}.",
                    f" Destination table contains {dest_row_count}.",
                )
            )
            return False

        logger.info("Source and destination table row counts match.")
        return True

    def create_table(self, source_table, destination_table):
        """
        Create the empty table in the destination database based on the source
        database schema structure. This method utilizes the Alchemy subclass.
        """

        # Try to create the destination using the source table's schema; if that doesn't work,
        # then we will lean on "copy" when loading the data to create the destination
        try:
            source_obj = self.source_db.get_table_object(source_table)
            self.dest_db.create_table(source_obj, destination_table)
        except Exception:
            logger.warning(
                "Unable to create destination table based on source table; we will "
                'fallback to using "copy" to create the destination.'
            )



import logging

from parsons.etl.table import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)


class Quickbase(object):
    """
    Instantiate the Quickbase class

    `Args:`
        hostname: str
            The URL for the homepage/login page of the organization's Quickbase
            instance (e.g. demo.quickbase.com).
        user_token: str
            The Quickbase account user token (API key). Not required if
            ``QUICKBASE_USER_TOKEN`` env variable is set.
    `Returns:`
        Quickbase Class
    """

    def __init__(self, hostname=None, user_token=None):
        self.hostname = check_env.check("QUICKBASE_HOSTNAME", hostname)
        self.user_token = check_env.check("QUICKBASE_USER_TOKEN", user_token)
        self.api_hostname = "https://api.quickbase.com/v1"
        self.client = APIConnector(
            self.api_hostname,
            headers={
                "QB-Realm-Hostname": self.hostname,
                "AUTHORIZATION": f"QB-USER-TOKEN {self.user_token}",
            },
        )

    def get_app_tables(self, app_id=None):
        """
        Query records in a Quickbase table. This follows the patterns laid out
        in Quickbase query documentaiton, located here:
        https://help.quickbase.com/api-guide/componentsquery.html

        `Args:`
            app_id: str
                Identifies which Quickbase app from which to fetch tables.
        `Returns:`
            Table Class
        """
        return Table(
            self.client.request(f"{self.api_hostname}/tables?appId={app_id}", "GET").json()
        )

    def query_records(self, table_from=None):
        """
        Query records in a Quickbase table. This follows the patterns laid out
        in Quickbase query documentaiton, located here:
        https://help.quickbase.com/api-guide/componentsquery.html

        `Args:`
            from: str
                The ID of a Quickbase resource (i.e. a table) to query.
        `Returns:`
            Table Class
        """
        req_resp = self.client.request(
            f"{self.api_hostname}/records/query", "POST", json={"from": table_from}
        ).json()

        resp_tbl = Table(req_resp["data"])
        cleaned_tbl = Table()

        for row in resp_tbl:
            row_dict = {}
            for column in resp_tbl.columns:
                row_dict[column] = row[column]["value"]
            cleaned_tbl.concat(Table([row_dict]))
            cleaned_tbl.materialize()

        column_resp = req_resp["fields"]
        column_map = {}
        for entry in column_resp:
            column_map[str(entry["id"])] = entry["label"].lower().strip()

        for column in cleaned_tbl.columns:
            cleaned_tbl.rename_column(column, column_map[column])

        return cleaned_tbl



import logging

from parsons.capitol_canary import CapitolCanary

logger = logging.getLogger(__name__)


class Phone2Action(object):
    """
    Instantiate Phone2Action Class

    `Args:`
        app_id: str
            The Phone2Action provided application id. Not required if ``PHONE2ACTION_APP_ID``
            env variable set.
        app_key: str
            The Phone2Action provided application key. Not required if ``PHONE2ACTION_APP_KEY``
            env variable set.
    `Returns:`
        Phone2Action Class
    """

    def __init__(self, app_id=None, app_key=None):
        self.capitol_canary = CapitolCanary(app_id, app_key)
        logger.warning("The Phone2Action class is being deprecated and replaced by CapitalCanary")

    def __getattr__(self, name):
        try:
            return getattr(self.capitol_canary, name)
        except AttributeError as e:
            raise AttributeError(f"{type(self).__name__} object has no attribute {name}") from e

    def get_advocates(self, state=None, campaign_id=None, updated_since=None, page=None):
        """
        Return advocates (person records).

        If no page is specified, the method will automatically paginate through the available
        advocates.

        `Args:`
            state: str
                Filter by US postal abbreviation for a state
                or territory e.g., "CA" "NY" or "DC"
            campaign_id: int
                Filter to specific campaign
            updated_since: str or int or datetime
                Fetch all advocates updated since the date provided; this can be a datetime
                object, a UNIX timestamp, or a date string (ex. '2014-01-05 23:59:43')
            page: int
                Page number of data to fetch; if this is specified, call will only return one
                page.
        `Returns:`
            A dict of parsons tables:
                * emails
                * phones
                * memberships
                * tags
                * ids
                * fields
                * advocates
        """
        return self.capitol_canary.get_advocates(state, campaign_id, updated_since, page)

    def get_campaigns(
        self,
        state=None,
        zip=None,
        include_generic=False,
        include_private=False,
        include_content=True,
    ):
        """
        Returns a list of campaigns

        `Args:`
            state: str
                Filter by US postal abbreviation for a state or territory e.g., "CA" "NY" or "DC"
            zip: int
                Filter by 5 digit zip code
            include_generic: boolean
                When filtering by state or ZIP code, include unrestricted campaigns
            include_private: boolean
                If true, will include private campaigns in results
            include_content: boolean
                If true, include campaign content fields, which may vary. This may cause
                sync errors.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        return self.capitol_canary.get_campaigns(
            state, zip, include_generic, include_private, include_content
        )

    def create_advocate(
        self,
        campaigns,
        first_name=None,
        last_name=None,
        email=None,
        phone=None,
        address1=None,
        address2=None,
        city=None,
        state=None,
        zip5=None,
        sms_optin=None,
        email_optin=None,
        sms_optout=None,
        email_optout=None,
        **kwargs,
    ):
        """
        Create an advocate.

        If you want to opt an advocate into or out of SMS / email campaigns, you must provide
        the email address or phone number (accordingly).

        The list of arguments only partially covers the fields that can be set on the advocate.
        For a complete list of fields that can be updated, see
        `the Phone2Action API documentation <https://docs.phone2action.com/#calls-create>`_.

        `Args:`
            campaigns: list
                The ID(s) of campaigns to add the advocate to
            first_name: str
                `Optional`; The first name of the advocate
            last_name: str
                `Optional`; The last name of the advocate
            email: str
                `Optional`; An email address to add for the advocate. One of ``email`` or ``phone``
                is required.
            phone: str
                `Optional`; An phone # to add for the advocate. One of ``email`` or ``phone`` is
                required.
            address1: str
                `Optional`; The first line of the advocates' address
            address2: str
                `Optional`; The second line of the advocates' address
            city: str
                `Optional`; The city of the advocates address
            state: str
                `Optional`; The state of the advocates address
            zip5: str
                `Optional`; The 5 digit Zip code of the advocate
            sms_optin: boolean
                `Optional`; Whether to opt the advocate into receiving text messages; an SMS
                confirmation text message will be sent. You must provide values for the ``phone``
                and ``campaigns`` arguments.
            email_optin: boolean
                `Optional`; Whether to opt the advocate into receiving emails. You must provide
                values for the ``email`` and ``campaigns`` arguments.
            sms_optout: boolean
                `Optional`; Whether to opt the advocate out of receiving text messages. You must
                provide values for the ``phone`` and ``campaigns`` arguments. Once an advocate is
                opted out, they cannot be opted back in.
            email_optout: boolean
                `Optional`; Whether to opt the advocate out of receiving emails. You must
                provide values for the ``email`` and ``campaigns`` arguments. Once an advocate is
                opted out, they cannot be opted back in.
            **kwargs:
                Additional fields on the advocate to update
        `Returns:`
            The int ID of the created advocate.
        """
        return self.capitol_canary.create_advocate(
            campaigns,
            first_name,
            last_name,
            email,
            phone,
            address1,
            address2,
            city,
            state,
            zip5,
            sms_optin,
            email_optin,
            sms_optout,
            email_optout,
            **kwargs,
        )

    def update_advocate(
        self,
        advocate_id,
        campaigns=None,
        email=None,
        phone=None,
        sms_optin=None,
        email_optin=None,
        sms_optout=None,
        email_optout=None,
        **kwargs,
    ):
        """
        Update the fields of an advocate.

        If you want to opt an advocate into or out of SMS / email campaigns, you must provide
        the email address or phone number along with a list of campaigns.

        The list of arguments only partially covers the fields that can be updated on the advocate.
        For a complete list of fields that can be updated, see
        `the Phone2Action API documentation <https://docs.phone2action.com/#calls-create>`_.

        `Args:`
            advocate_id: integer
                The ID of the advocate being updates
            campaigns: list
                `Optional`; The ID(s) of campaigns to add the user to
            email: str
                `Optional`; An email address to add for the advocate (or to use when opting in/out)
            phone: str
                `Optional`; An phone # to add for the advocate (or to use when opting in/out)
            sms_optin: boolean
                `Optional`; Whether to opt the advocate into receiving text messages; an SMS
                confirmation text message will be sent. You must provide values for the ``phone``
                and ``campaigns`` arguments.
            email_optin: boolean
                `Optional`; Whether to opt the advocate into receiving emails. You must provide
                values for the ``email`` and ``campaigns`` arguments.
            sms_optout: boolean
                `Optional`; Whether to opt the advocate out of receiving text messages. You must
                provide values for the ``phone`` and ``campaigns`` arguments. Once an advocate is
                opted out, they cannot be opted back in.
            email_optout: boolean
                `Optional`; Whether to opt the advocate out of receiving emails. You must
                provide values for the ``email`` and ``campaigns`` arguments. Once an advocate is
                opted out, they cannot be opted back in.
            **kwargs:
                Additional fields on the advocate to update
        """
        return self.capitol_canary.update_advocate(
            advocate_id,
            campaigns,
            email,
            phone,
            sms_optin,
            email_optin,
            sms_optout,
            email_optout,
            **kwargs,
        )



"""Pydantic data models for use with dbt utilities."""

import collections

from dbt.contracts.graph.manifest import Manifest as dbtManifest
from dbt.contracts.results import NodeResult, SourceFreshnessResult


class Manifest:
    def __init__(self, command: str, dbt_manifest: dbtManifest) -> None:
        self.command = command
        self.dbt_manifest = dbt_manifest

    def __getattr__(self, key):
        if key in dir(self):
            result = getattr(self, key)
        elif (
            getattr(self.dbt_manifest, "metadata", {})
            and key in self.dbt_manifest.metadata.__dict__
        ):
            result = getattr(self.dbt_manifest.metadata, key)
        else:
            result = getattr(self.dbt_manifest, key)
        return result

    def filter_results(self, **kwargs) -> list[NodeResult]:
        """Subset of results based on filter"""
        filtered_results = [
            result
            for result in self.dbt_manifest
            if all(str(getattr(result, key)) == value for key, value in kwargs.items())
        ]
        return filtered_results

    @property
    def warnings(self) -> list[NodeResult]:
        return self.filter_results(status="warn")

    @property
    def errors(self) -> list[NodeResult]:
        return self.filter_results(status="error")

    @property
    def fails(self) -> list[NodeResult]:
        return self.filter_results(status="fail")

    @property
    def skips(self) -> list[NodeResult]:
        """Returns skipped model builds but not skipped tests."""
        return [
            node
            for node in self.filter_results(status="skipped")
            if node.node.name.split(".")[0] == "model"
        ]

    @property
    def summary(self) -> collections.Counter:
        """Counts of pass, warn, fail, error & skip."""
        result = collections.Counter([str(i.status) for i in self.dbt_manifest])
        return result

    @property
    def total_gb_processed(self) -> float:
        """Total GB processed by full dbt command run."""
        result = (
            sum([node.adapter_response.get("bytes_processed", 0) for node in self.dbt_manifest])
            / 1000000000
        )
        return result

    @property
    def total_slot_hours(self) -> float:
        """Total slot hours used by full dbt command run."""
        result = (
            sum([node.adapter_response.get("slot_ms", 0) for node in self.dbt_manifest]) / 3600000
        )
        return result


class EnhancedNodeResult(NodeResult):
    def log_message(self) -> str | None:
        """Helper method to generate message for logs."""
        if isinstance(self, SourceFreshnessResult):
            freshness_config = self.node.freshness
            time_config = getattr(freshness_config, self.status + "_after")
            result = f"No new records for {int(self.age / 86400)} days, {self.status} after {time_config.count} {time_config.period}s."
        else:
            result = self.message

        return result



"""NGPVAN Saved List Endpoints"""

import logging

from parsons.etl.table import Table

logger = logging.getLogger(__name__)


class PrintedLists(object):
    def __init__(self, van_connection):
        self.connection = van_connection

    def get_printed_lists(
        self,
        generated_after=None,
        generated_before=None,
        created_by=None,
        folder_name=None,
        turf_name=None,
    ):
        """
        Get printed lists.

        `Args:`
            folder_id: int
                Filter by the id for a VAN folder. If included returns only
                the saved lists in the folder
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {
            "generatedAfter": generated_after,
            "generatedBefore": generated_before,
            "createdBy": created_by,
            "folderName": folder_name,
            "turfName": turf_name,
        }

        params = {key: value for key, value in params.items() if value is not None}

        tbl = Table(self.connection.get_request("printedLists", params=params))

        logger.info(f"Found {tbl.num_rows} printed lists.")
        return tbl

    def get_printed_list(self, printed_list_number):
        """
        Returns a printed list object.

        `Args:`
            printed_list_number: int
                The printed list number
        `Returns:`
            dict
        """

        r = self.connection.get_request(f"printedLists/{printed_list_number}")
        logger.info(f"Found printed list {printed_list_number}.")
        return r



import os
import unittest
from copy import deepcopy

import requests_mock

from parsons import VAN, Table


def assert_matching_tables(table1, table2, ignore_headers=False):
    if ignore_headers:
        data1 = table1.data
        data2 = table2.data
    else:
        data1 = table1
        data2 = table2

    if isinstance(data1, Table) and isinstance(data2, Table):
        assert data1.num_rows == data2.num_rows

    for r1, r2 in zip(data1, data2):
        # Cast both rows to lists, in case they are different types of collections. Must call
        # .items() on dicts to compare content of collections
        if isinstance(r1, dict):
            r1 = r1.items()
        if isinstance(r2, dict):
            r2 = r2.items()

        assert list(r1) == list(r2)


os.environ["VAN_API_KEY"] = "SOME_KEY"


sample_content_full = [
    {
        "name": "Email Name",
        "senderDisplayName": "Joe Biden",
        "senderEmailAddress": "joe@biden.org",
        "subject": "This is Joe",
        "createdBy": "Random Intern",
        "dateCreated": "2023-05-17T15:04:00Z",
        "emailMessageContentDistributions": {
            "dateSent": "2023-05-17T15:05:00Z",
            "recipientCount": 10,
            "openCount": 0,
            "linksClickedCount": 0,
            "unsubscribeCount": 0,
            "bounceCount": 0,
            "contributionTotal": 0.0,
            "formSubmissionCount": 0,
            "contributionCount": 0,
            "machineOpenCount": 0,
        },
    },
    {
        "name": "Email Name_B",
        "senderDisplayName": "Kamala Harris",
        "senderEmailAddress": "kamala@harris.org",
        "subject": "This is Kamala",
        "createdBy": "Random Intern",
        "dateCreated": "2023-05-17T15:04:00Z",
        "emailMessageContentDistributions": {
            "dateSent": "2023-05-17T15:05:00Z",
            "recipientCount": 10,
            "openCount": 1,
            "linksClickedCount": 1,
            "unsubscribeCount": 1,
            "bounceCount": 0,
            "contributionTotal": 1.0,
            "formSubmissionCount": 0,
            "contributionCount": 1,
            "machineOpenCount": 1,
        },
    },
    {
        "name": "Email Name_C",
        "senderDisplayName": "Commander Biden",
        "senderEmailAddress": "commander@barkbark.org",
        "subject": "Bark Bark",
        "createdBy": "Random Intern",
        "dateCreated": "2023-05-17T15:04:00Z",
        "emailMessageContentDistributions": {
            "dateSent": "2023-05-17T15:05:00Z",
            "recipientCount": 10,
            "openCount": 5,
            "linksClickedCount": 0,
            "unsubscribeCount": 0,
            "bounceCount": 0,
            "contributionTotal": 0.0,
            "formSubmissionCount": 0,
            "contributionCount": 0,
            "machineOpenCount": 2,
        },
    },
    {
        "name": "Email Name_Winner",
        "senderDisplayName": "Commander Biden",
        "senderEmailAddress": "commander@barkbark.org",
        "subject": "Bark Bark",
        "createdBy": "TargetedEmail InternalAPIUser",
        "dateCreated": "2023-05-17T17:06:00Z",
        "emailMessageContentDistributions": {
            "dateSent": "2023-05-17T17:06:00Z",
            "recipientCount": 100,
            "openCount": 30,
            "linksClickedCount": 10,
            "unsubscribeCount": 10,
            "bounceCount": 10,
            "contributionTotal": 500.0,
            "formSubmissionCount": 0,
            "contributionCount": 3,
            "machineOpenCount": 30,
        },
    },
]

sample_content_single = [
    {
        "name": "Email Name",
        "senderDisplayName": "Joe Biden",
        "senderEmailAddress": "joe@biden.org",
        "subject": "This is Joe",
        "createdBy": "Random Intern",
        "dateCreated": "2023-05-17T15:04:00Z",
        "emailMessageContentDistributions": {
            "dateSent": "2023-05-17T15:05:00Z",
            "recipientCount": 10,
            "openCount": 0,
            "linksClickedCount": 0,
            "unsubscribeCount": 0,
            "bounceCount": 0,
            "contributionTotal": 0.0,
            "formSubmissionCount": 0,
            "contributionCount": 0,
            "machineOpenCount": 0,
        },
    }
]

sample_content_pending_email = [
    {
        "name": "I'm a pending email",
        "senderDisplayName": None,
        "senderEmailAddress": None,
        "subject": None,
        "createdBy": "Random Intern",
        "dateCreated": "2023-05-17T15:04:00Z",
        "emailMessageContentDistributions": None,
    }
]

mock_response = [
    {
        "foreignMessageId": "oK2ahdAcEe6F-QAiSCI3lA2",
        "name": "Test Email",
        "createdBy": "Joe Biden",
        "dateCreated": "2024-02-20T13:20:00Z",
        "dateScheduled": "2024-02-20T15:55:00Z",
        "campaignID": 0,
        "dateModified": "2024-02-20T15:54:36.27Z",
        "emailMessageContent": None,
    },
    {
        "foreignMessageId": "rjzc2szzEe6F-QAiSCI3lA2",
        "name": "Test Email 2",
        "createdBy": "Joe Biden",
        "dateCreated": "2024-02-16T12:49:00Z",
        "dateScheduled": "2024-02-16T13:29:00Z",
        "campaignID": 0,
        "dateModified": "2024-02-16T13:29:16.453Z",
        "emailMessageContent": None,
    },
    {
        "foreignMessageId": "E1AfcnkEe6F-QAiSCI3lA2",
        "name": "Test Email 3",
        "createdBy": "Joe Biden",
        "dateCreated": "2024-02-12T15:26:00Z",
        "dateScheduled": "2024-02-13T11:22:00Z",
        "campaignID": 0,
        "dateModified": "2024-02-13T11:22:28.273Z",
        "emailMessageContent": None,
    },
    {
        "foreignMessageId": "6GTLBsUwEe62YAAiSCIxlw2",
        "name": "Test Email 4",
        "createdBy": "Joe Biden",
        "dateCreated": "2024-02-06T15:47:00Z",
        "dateScheduled": "2024-02-07T10:32:00Z",
        "campaignID": 0,
        "dateModified": "2024-02-07T10:31:55.16Z",
        "emailMessageContent": None,
    },
    {
        "foreignMessageId": "mgTdmcEiEe62YAAiSCIxlw2",
        "name": "Test Email 5",
        "createdBy": "Joe Biden",
        "dateCreated": "2024-02-01T11:55:00Z",
        "dateScheduled": "2024-02-01T16:08:00Z",
        "campaignID": 0,
        "dateModified": "2024-02-01T16:08:10.737Z",
        "emailMessageContent": None,
    },
]

mock_response_enriched = deepcopy(mock_response)
mock_response_enriched[0]["emailMessageContent"] = sample_content_single
mock_response_enriched[1]["emailMessageContent"] = sample_content_single
mock_response_enriched[2]["emailMessageContent"] = sample_content_single
mock_response_enriched[3]["emailMessageContent"] = sample_content_pending_email
mock_response_enriched[4]["emailMessageContent"] = sample_content_full


class TestEmail(unittest.TestCase):
    def setUp(self):
        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters")

    @requests_mock.Mocker()
    def test_get_email_messages(self, m):
        m.get(
            self.van.connection.uri + "email/messages",
            json=mock_response,
            status_code=200,
        )

        response = self.van.get_emails()

        assert_matching_tables(response, mock_response)

    @requests_mock.Mocker()
    def test_get_email_message(self, m):
        m.get(
            self.van.connection.uri + "email/message/0",
            json=mock_response[0],
            status_code=200,
        )
        m.get(
            self.van.connection.uri + "email/message/1",
            json=mock_response_enriched[0],
            status_code=200,
        )
        response0 = self.van.get_email(0, expand=False)
        response1 = self.van.get_email(1)

        assert_matching_tables(response0, mock_response[0])
        assert_matching_tables(response1, mock_response_enriched[0])

    @requests_mock.Mocker()
    def test_get_email_message_stats_agg(self, m):
        m.get(
            self.van.connection.uri + "email/messages",
            json=mock_response_enriched,
            status_code=200,
        )
        for resp in mock_response_enriched:
            m.get(
                self.van.connection.uri + f"email/message/{resp['foreignMessageId']}",
                json=resp,
                status_code=200,
            )

        response_t = self.van.get_email_stats(aggregate_ab=True)
        response_f = self.van.get_email_stats(aggregate_ab=False)

        assert len(response_t.to_dicts()) == 5
        assert len(response_f.to_dicts()) == 8

        assert response_t.to_dicts()[4]["recipientCount"] == 130
        assert response_f.to_dicts()[4]["recipientCount"] == 10



# Dictionaries used to map column names to valid argument names

LEAD_COLUMN_MAP = {
    "first_name": ["first", "fn"],
    "last_name": ["last", "ln", "lastname"],
    "phone_number": ["phone", "cell", "cellphone"],
    "email": ["emailaddress"],
    "follow_up": ["follow"],
}



import csv
import typing as t
import zipfile
from dataclasses import dataclass
from datetime import datetime
from io import BytesIO, StringIO

import defusedxml.ElementTree as ET
import requests
from dateutil.parser import parse as parsedate
from pytz import timezone

CLARITY_URL = "https://results.enr.clarityelections.com/"

CURRENT_VERSION_URL_TEMPLATE = CLARITY_URL + "{administrator}/{election_id}/current_ver.txt"
SUMMARY_CSV_ZIP_URL_TEMPLATE = (
    CLARITY_URL + "{administrator}/{election_id}/{version_num}/reports/summary.zip"
)
DETAIL_XML_ZIP_URL_TEMPLATE = (
    CLARITY_URL + "{administrator}/{election_id}/{version_num}/reports/detailxml.zip"
)
COUNTY_DETAIL_XML_ZIP_URL_TEMPLATE = (
    CLARITY_URL
    + "{state}/{county_name}/{county_election_id}/{county_version_num}/reports/detailxml.zip"
)
ELECTION_SETTINGS_JSON_URL_TEMPLATE = (
    CLARITY_URL + "{state}/{election_id}/{version_num}/json/en/electionsettings.json"
)

BROWSER_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) "
    "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36"
}

TZ_INFO = {
    "EST": "UTC-5",
    "EDT": "UTC-4",
    "CST": "UTC-6",
    "CDT": "UTC-5",
    "MST": "UTC-7",
    "MDT": "UTC-6",
    "PST": "UTC-8",
    "PDT": "UTC-7",
    "AKST": "UTC-9",
    "AKDT": "UTC-8",
    "HST": "UTC-10",
    "HDT": "UTC-9",
}


@dataclass
class CountyDetails:
    """
    A class for keeping track of County election details.

    A dataclass is decorator that adds special functions including an
    automatic __init__ function. See more here: https://docs.python.org/3/library/dataclasses.html
    """

    state: str
    county_name: str
    county_election_id: str
    county_version_num: str
    county_update_date: datetime = None


class Scytl:
    """
    Instantiate a Scytl connector.

    `Args:`:
        state: str
            The two letter code of the state the publishing election results.
            ex: GA
        election_id: str
            The numeric identifier for the election found in the url of the election's website.
            ex: "114729"
        county: str (optional)
            The name of the county publishing the results.
            ex: Clarke
    """

    def __init__(self, state: str, election_id: str, county=""):
        self.state = state
        self.county = county.replace(" ", "_")

        self.administrator = f"{self.state}/{self.county}" if self.county else self.state
        self.election_id = election_id

        self.previous_summary_version_num = None
        self.previous_details_version_num = None
        self.previous_county_details_version_num = None
        self.previous_county_details_list = None
        self.previously_fetched_counties = set()

    def _parse_date_to_utc(self, input_dt: str) -> datetime:
        """
        Parse datetime string as datetime in UTC

        `Args`:
            input_dt: str
                The datetime string to be parsed
        `Returns`:
            datetime | None
        """

        if input_dt is None:
            return

        temp = parsedate(input_dt, tzinfos=TZ_INFO)
        temp = temp.astimezone(timezone("UTC"))

        return temp

    def _get_version(self, administrator: str, election_id: str) -> str:
        """
        Fetch the latest version of the election results from the Clarity site

        `Args`:
            administrator: str
                The url code for the election administrator, either the two-letter
                state code or the state code and the county, separated by a slash
            election_id: str
                The election id for the given election as a string
        `Returns`:
            str
            The version id as a string
        """

        config_version_url = CURRENT_VERSION_URL_TEMPLATE.format(
            administrator=administrator, election_id=election_id
        )

        res = requests.get(config_version_url, headers=BROWSER_HEADERS)

        return res.text

    def _parse_file_from_zip_url(self, zipfile_url: str, file_name: str) -> bytes:
        """
        Fetch a zip file from the given url and unzip to a byte array

        `Args`:
            zipfile_url: str
                The url where the zip file can be found
            election_id: str
                The expected name of the file in the zipfile to read
        `Returns`:
            bytes
            The unzipped file as bytes
        """

        with BytesIO() as zipdata:
            with requests.get(zipfile_url, headers=BROWSER_HEADERS) as res:
                zipdata.write(res.content)
                zipdata.flush()

            zf = zipfile.ZipFile(zipdata)

            with zf.open(file_name) as input:
                return input.read()

    def _get_latest_counties_scytl_info(
        self, state: str, election_id: str, version_num: str
    ) -> t.Dict[str, CountyDetails]:
        """
        Fetch the settings JSON file for the election and parse the county details
        for participating counties in a state election.

        `Args`:
            state: str
                The two-letter state code for the state
            election_id: str
                The election ID for the given election
            version_num: str
                The latest version ID of the election as a string
        `Returns`:
            dict[str, CountyDetails]
            A dictionary mapping county names to their sub-election information
        """

        county_dict = {}

        config_settings_json_url = ELECTION_SETTINGS_JSON_URL_TEMPLATE.format(
            state=state, election_id=election_id, version_num=version_num
        )

        settings_json_res = requests.get(config_settings_json_url, headers=BROWSER_HEADERS)
        settings_json = settings_json_res.json()

        participating_counties = settings_json["settings"]["electiondetails"][
            "participatingcounties"
        ]

        for county_row in participating_counties:
            county_info = county_row.split("|")
            source_county_name = county_info[0]
            county_election_id = county_info[1]
            county_version_num = county_info[2]
            county_update_date = self._parse_date_to_utc(county_info[3])

            county_details = CountyDetails(
                state,
                source_county_name,
                county_election_id,
                county_version_num,
                county_update_date,
            )

            county_dict[source_county_name] = county_details

        return county_dict

    def _parse_county_xml_data_to_precincts(
        self, county_data: bytes, county_details: CountyDetails
    ) -> t.List[t.Dict]:
        """
        Parse a detail XML file for a county into a list of election
        results by precinct and vote method.

        `Args`:
            county_data: bytes
                The detail XML file for a county as bytes
            county_details: str
                The details class for the county, including name,
                id, and last updated datetime
        `Returns`:
            list[dict]
            The list of election results by precinct and vote method in the file.
        """

        tree = ET.fromstring(county_data)

        precinct_dict = {}
        precinct_votes = []

        root = tree

        for child in root:
            if child.tag == "VoterTurnout":
                precincts = child[0]

                for precinct in precincts:
                    data = precinct.attrib
                    name = data.get("name")

                    precinct_info = {
                        "total_voters": data.get("totalVoters"),
                        "ballots_cast": data.get("ballotsCast"),
                        "voter_turnout": data.get("voterTurnout"),
                        "percent_reporting": data.get("percentReporting"),
                    }

                    precinct_dict[name] = precinct_info

            if child.tag == "Contest":
                office = child.attrib["text"]

                for choice in child:
                    cand_votes = {}

                    if choice.tag == "VoteType":
                        continue

                    source_cand_data = choice.attrib
                    cand_name = source_cand_data.get("text")
                    cand_party = source_cand_data.get("party")

                    for vote_type in choice:
                        vote_type_label = vote_type.attrib["name"]

                        for precinct in vote_type:
                            precinct_name = precinct.attrib["name"]
                            cand_votes[precinct_name] = int(precinct.attrib["votes"])

                            precinct_turnout = precinct_dict.get(precinct_name, {})

                            result = {
                                "state": county_details.state,
                                "county_name": county_details.county_name,
                                "county_id": county_details.county_election_id,
                                "office": office,
                                "ballots_cast": precinct_turnout.get("ballots_cast"),
                                "reg_voters": precinct_turnout.get("total_voters"),
                                "vote_method": vote_type_label,
                                "candidate_name": cand_name,
                                "candidate_party": cand_party,
                                "precinct_name": precinct_name,
                                "recorded_votes": cand_votes[precinct_name],
                                "voter_turnout": precinct_turnout.get("voter_turnout"),
                                "percent_reporting": precinct_turnout.get("percent_reporting"),
                                "timestamp_last_updated": county_details.county_update_date,
                            }

                            precinct_votes.append(result)

        return precinct_votes

    def _parse_state_xml_data_to_counties(self, state_data: bytes, state: str) -> t.List[t.Dict]:
        """
        Parse a detail XML file for a state into a list of election
        results by county and vote method.

        `Args`:
            state_data: bytes
                The detail XML file for a state as bytes
            state: str
                The two-letter state code for the state associated with the file
        `Returns`:
            list[dict]
            The list of election results by state and vote method in the file.
        """

        root = ET.fromstring(state_data)

        county_dict = {}
        county_votes = []

        timestamp = None

        for child in root:
            if child.tag == "Timestamp":  # <Timestamp>1/5/2021 3:22:30 PM EST</Timestamp>
                timestamp = self._parse_date_to_utc(child.text)

            if child.tag == "ElectionVoterTurnout":
                counties = child[0]

                for county in counties:
                    data = county.attrib
                    name = data["name"]

                    county_dict[name] = data

            if child.tag == "Contest":
                office = child.attrib["text"]

                for choice in child:
                    cand_votes = {}

                    if choice.tag == "ParticipatingCounties":
                        continue

                    source_cand_data = choice.attrib
                    cand_name = source_cand_data.get("text")
                    cand_party = source_cand_data.get("party")

                    for vote_type in choice:
                        vote_type_label = vote_type.attrib["name"]

                        for county in vote_type:
                            county_name = county.attrib["name"]
                            cand_votes[county_name] = int(county.attrib["votes"])

                            county_turnout = county_dict.get(county_name, {})

                            result = {
                                "state": state,
                                "county_name": county_name,
                                "office": office,
                                "ballots_cast": county_turnout.get("ballotsCast"),
                                "reg_voters": county_turnout.get("totalVoters"),
                                "precincts_reporting": county_turnout.get("precinctsReported"),
                                "total_precincts": county_turnout.get("precinctsParticipating"),
                                "vote_method": vote_type_label,
                                "candidate_name": cand_name,
                                "candidate_party": cand_party,
                                "recorded_votes": cand_votes[county_name],
                                "timestamp_last_updated": timestamp,
                            }

                            county_votes.append(result)

        return county_votes

    def _fetch_and_parse_summary_results(
        self, administrator: str, election_id: str, version_num: str, county=""
    ) -> t.List[t.Dict]:
        """
        Fetches the summary results CSV file from the Scytl site and parses it
        into a list of election results by candidate.

        `Args`:
            administrator: str
                The url code for the election administrator, either the two-letter
                state code or the state code and the county, separated by a slash
            election_id: str
                The election id for the given election as a string
            version_num: str
                The latest version ID of the election as a string
            county: str
                The name of the county associated with the summary file
        `Returns`:
            list[dict]
            The list of election results by candidate.
        """

        summary_csv_zip_url = SUMMARY_CSV_ZIP_URL_TEMPLATE.format(
            administrator=administrator,
            election_id=election_id,
            version_num=version_num,
        )

        zip_bytes = self._parse_file_from_zip_url(summary_csv_zip_url, "summary.csv")

        string_buffer = StringIO(zip_bytes.decode("latin-1"))
        csv_data = csv.DictReader(string_buffer, delimiter=",")

        data = [
            {
                "state": self.state,
                "county_name": county or self.county,
                "office": x.get("contest name"),
                "ballots_cast": x.get("ballots cast"),
                "reg_voters": x.get("registered voters"),
                "counties_reporting": x.get("num Area rptg"),
                "total_counties": x.get("num Area total"),
                "precincts_reporting": x.get("num Precinct rptg"),
                "total_precincts": x.get("num Precinct total"),
                "candidate_name": x.get("choice name"),
                "candidate_party": x.get("party name"),
                "recorded_votes": x.get("total votes"),
            }
            for x in csv_data
        ]

        return data

    def get_summary_results(self, force_update=False) -> t.List[t.Dict]:
        """
        Fetch the latest summary results for the given election, across all contests.

        Please note that all electoral entities administer their elections differently,
            so not all values will be populated if the entity doesn't provide them.

        `Args:`
            force_update: bool
                If this is False, the connector will check to see if the current version
                    matches the previously fetched version of the results.
                    If the version has not been changed, no results will be fetched or returned.
                Default: false
        `Returns:`
            list[dict]
            The list should contain entries for each candidate in each office.
            Each row will contain the following:
            - state
            - county_name (if applicable)
            - office
            - ballots_cast (in the contest)
            - reg_voters (eligible for the contest)
            - counties_reporting
            - total_counties
            - precincts_reporting
            - total_precincts
            - candidate_name
            - candidate_party (many administrators do not use this feature
                and instead include the party in the candidate name)
            - recorded_votes (votes cast for the candidate)
        """

        version_num = self._get_version(self.administrator, self.election_id)

        if not force_update and version_num == self.previous_summary_version_num:
            return

        data = self._fetch_and_parse_summary_results(
            self.administrator, self.election_id, version_num
        )

        self.previous_summary_version_num = version_num

        return data

    def get_detailed_results(self, force_update=False) -> t.List[t.Dict]:
        """
        Fetch the latest detailed results by geography for the given election, across all contests.

        Please note that all electoral entities administer their elections differently,
            so not all values will be populated if the entity doesn't provide them.

        `Args:`
            force_update: bool
                If this is False, the connector will check to see if the current version
                    matches the previously fetched version of the results.
                    If the version has not been changed, no results will be fetched or returned.
                Default: false
        `Returns:`
            list[dict]
            The list should contain entries for each candidate in each office,
                per vote method and per county.

            If fetching for a state, results will look like:
            - state
            - county_name
            - office
            - ballots_cast
            - reg_voters
            - precincts_reporting
            - total_precincts
            - vote_method (note: some administrators choose to differentiate
                results by vote method, while others do not)
            - candidate_name
            - candidate_party (many administrators do not use this
                feature and instead include the party in the candidate name)
            - recorded_votes (votes cast for the candidate
                with this vote method in this county)
            - timestamp_last_updated

            If fetching for a county, results will look like:
            - state
            - county_name
            - county_id
            - office
            - ballots_cast
            - reg_voters
            - vote_method (note: some administrators choose to
                differentiate results by vote method, while others do not)
            - candidate_name
            - candidate_party (many administrators do not use this
                feature and instead include the party in the candidate name)
            - precinct_name
            - recorded_votes (votes cast for the candidate
                with this vote method in this county)
            - voter_turnout
            - percent_reporting
            - timestamp_last_updated
        """

        version_num = self._get_version(self.administrator, self.election_id)

        if not force_update and version_num == self.previous_details_version_num:
            return

        detail_xml_url = DETAIL_XML_ZIP_URL_TEMPLATE.format(
            administrator=self.administrator,
            election_id=self.election_id,
            version_num=version_num,
        )

        parsed_data = []

        county_data = self._parse_file_from_zip_url(detail_xml_url, "detail.xml")

        if self.county:
            county_details = CountyDetails(self.state, self.county, self.election_id, version_num)

            parsed_data = self._parse_county_xml_data_to_precincts(county_data, county_details)
        else:
            parsed_data = self._parse_state_xml_data_to_counties(county_data, self.state)

        self.previous_details_version_num = version_num

        return parsed_data

    def get_detailed_results_for_participating_counties(
        self, county_names: t.List[str] = None, force_update=False
    ) -> t.Tuple[t.List[str], t.List[t.Dict]]:
        """
        Fetch the latest detailed results for the given election for all participating counties
            with detailed results, across all contests.

        Some counties may not have detailed results. If so, this will attempt
            to fetch the summary results for that county. If no results exist for either,
            the county name will be appended to the missing_counties list.

        After the first fetch, only the counties with updates will be returned,
            previous results will not be included.

        Please note that all electoral entities administer their elections differently,
            so not all values will be populated if the entity doesn't provide them.

        `Args:`
            county_names: list[str]
                The list of counties to get precinct-level results for.
                Default: None (get all counties)
            force_update: bool
                If this is False, the connector will check to see if the current
                    version matches the previously fetched version of the results.
                    If the version has not been changed, no results will be fetched or returned.
                Default: false

        `Returns:`
            list[str]
            The list of county names that could not be fetched

            list[dict]
            The list should contain entries for each candidate in
                each office, per vote method, county, and precinct.
            Each row will contain the following:
            - state
            - county_name
            - county_id
            - office
            - ballots_cast
            - reg_voters
            - vote_method (note: some administrators choose to differentiate
                results by vote method, while others do not)
            - candidate_name
            - candidate_party (many administrators do not use this feature
                and instead include the party in the candidate name)
            - precinct_name
            - recorded_votes (votes cast for the candidate with this vote method in this county)
            - voter_turnout
            - percent_reporting
            - timestamp_last_updated
        """

        version_num = self._get_version(self.administrator, self.election_id)

        if not force_update and version_num == self.previous_county_details_version_num:
            return [], []

        county_details_list = self._get_latest_counties_scytl_info(
            self.state, self.election_id, version_num
        )

        parsed_data = []
        fetched_counties = []
        missing_counties = []

        for county_name, county_details in county_details_list.items():
            if county_names and county_name not in county_names:
                continue

            if (
                not force_update
                and county_name in self.previously_fetched_counties
                and self.previous_county_details_list
                and county_details.county_update_date
                <= self.previous_county_details_list[county_name].county_update_date
            ):
                continue

            detail_xml_url = COUNTY_DETAIL_XML_ZIP_URL_TEMPLATE.format(
                state=county_details.state,
                county_name=county_details.county_name,
                county_election_id=county_details.county_election_id,
                county_version_num=county_details.county_version_num,
            )

            try:
                county_data = self._parse_file_from_zip_url(detail_xml_url, "detail.xml")

            except requests.exceptions.RequestException:
                try:
                    summary_data = self._fetch_and_parse_summary_results(
                        f"{self.state}/{county_name}",
                        county_details.county_election_id,
                        county_details.county_version_num,
                        county_name,
                    )

                except requests.exceptions.RequestException:
                    missing_counties.append(county_name)

                else:
                    if len(summary_data) > 0:
                        parsed_data += summary_data

            else:
                parsed_data += self._parse_county_xml_data_to_precincts(county_data, county_details)

                fetched_counties.append(county_name)

        self.previous_county_details_version_num = version_num
        self.previous_county_details_list = county_details_list
        self.previously_fetched_counties = set(fetched_counties)

        return missing_counties, parsed_data



import logging
from typing import Dict, List, Union

from parsons.utilities import json_format

logger = logging.getLogger(__name__)


class People(object):
    def __init__(self, van_connection):
        self.connection = van_connection

    def find_person(
        self,
        first_name=None,
        last_name=None,
        date_of_birth=None,
        email=None,
        phone=None,
        phone_type=None,
        street_number=None,
        street_name=None,
        zip=None,
        **kwargs,
    ):
        """
        Find a person record.

        .. note::
            Person find must include the following minimum combinations to conduct
            a search.

              - first_name, last_name, email
              - first_name, last_name, phone
              - first_name, last_name, zip5, date_of_birth
              - first_name, last_name, street_number, street_name, zip5
              - email_address

        `Args:`
            first_name: str
                The person's first name
            last_name: str
                The person's last name
            dob: str
                ISO 8601 formatted date of birth (e.g. ``1981-02-01``)
            email: str
                The person's email address
            phone: str
                Phone number of any type (Work, Cell, Home)
            street_number: str
                Street Number
            street_name: str
                Street Name
            zip: str
                5 digit zip code
            kwargs:
                Any additional keyword arguments will be passed to
                the EveryAction API for matching.
        `Returns:`
            A person dict object
        """

        logger.info(f"Finding {first_name} {last_name}.")

        return self._people_search(
            first_name=first_name,
            last_name=last_name,
            date_of_birth=date_of_birth,
            email=email,
            phone=phone,
            phone_type=phone_type,
            street_number=street_number,
            street_name=street_name,
            zip=zip,
            **kwargs,
        )

    def find_person_json(self, match_json):
        """
        Find a person record based on json data.

        .. note::
            Person find must include the following minimum combinations to conduct
            a search.

              - first_name, last_name, email
              - first_name, last_name, phone
              - first_name, last_name, zip5, date_of_birth
              - first_name, last_name, street_number, street_name, zip5
              - email_address

        .. note::
            A full list of possible values for the json, and its structure can be found
            `here <https://docs.ngpvan.com/reference/people#peoplefind>`_.

        `Args:`
            match_json: dict
                A dictionary of values to match against.
            fields: The fields to return. Leave as default for all available fields
        `Returns:`
            A person dict object
        """

        logger.info("Finding a match for json details.")

        return self._people_search(match_json=match_json)

    def update_person(
        self,
        id=None,
        id_type="vanid",
        first_name=None,
        last_name=None,
        date_of_birth=None,
        email=None,
        phone=None,
        phone_type=None,
        street_number=None,
        street_name=None,
        zip=None,
    ):
        """
        Update a person record based on a provided ID. All other arguments provided will be
        updated on the record.

        .. warning::
            This method can only be run on MyMembers, EveryAction, MyCampaign databases.

        `Args:`
            id: str
                A valid id
            id_type: str
                A known person identifier type available on this VAN instance.
                Defaults to ``vanid``.
            first_name: str
                The person's first name
            last_name: str
                The person's last name
            dob: str
                ISO 8601 formatted date of birth (e.g. ``1981-02-01``)
            email: str
                The person's email address
            phone: str
                Phone number of any type (Work, Cell, Home)
            phone_type: str
                One of 'H' for home phone, 'W' for work phone, 'C' for cell, 'M' for
                main phone or 'F' for fax line. Defaults to home phone.
            street_number: str
                Street Number
            street_name: str
                Street Name
            zip: str
                5 digit zip code
        `Returns:`
            A person dict
        """

        return self._people_search(
            id=id,
            id_type=id_type,
            first_name=first_name,
            last_name=last_name,
            date_of_birth=date_of_birth,
            email=email,
            phone=phone,
            phone_type=phone_type,
            street_number=street_number,
            street_name=street_name,
            zip=zip,
            create=True,
        )

    def update_person_json(self, id, id_type="vanid", match_json=None):
        """
        Update a person record based on a provided ID within the match_json dict.

        .. note::
            A full list of possible values for the json, and its structure can be found
            `here <https://docs.ngpvan.com/reference/people#peoplevanid>`_.

        `Args:`
            id: str
                A valid id
            id_type: str
                A known person identifier type available on this VAN instance.
                Defaults to ``vanid``.
            match_json: dict
                A dictionary of values to match against and save.
        `Returns:`
            A person dict
        """

        return self._people_search(id=id, id_type=id_type, match_json=match_json, create=True)

    def upsert_person(
        self,
        first_name=None,
        last_name=None,
        date_of_birth=None,
        email: Union[str, List[Dict[str, Union[str, bool]]], None] = None,
        phone=None,
        phone_type=None,
        street_number=None,
        street_name=None,
        zip=None,
        **kwargs,
    ):
        """
        Create or update a person record.

        .. note::
            Person find must include the following minimum combinations.

              - first_name, last_name, email
              - first_name, last_name, phone
              - first_name, last_name, zip5, date_of_birth
              - first_name, last_name, street_number, street_name, zip5
              - email_address

        .. warning::
            This method can only be run on MyMembers, EveryAction, MyCampaign databases.

        `Args:`
            first_name: str
                The person's first name
            last_name: str
                The person's last name
            dob: str
                ISO 8601 formatted date of birth (e.g. ``1981-02-01``)
            email: Union[str, List[Dict[str, Union[str, bool]]], None]
                The person's email address or a list of email dicts.
                e.g. [{'email': 'abcd@gmail.com', 'isSubscribed': False}]
                See https://docs.everyaction.com/reference/people-common-models#email
            phone: str
                Phone number of any type (Work, Cell, Home)
            phone_type: str
                One of 'H' for home phone, 'W' for work phone, 'C' for cell, 'M' for
                main phone or 'F' for fax line. Defaults to home phone.
            street_number: str
                Street Number
            street_name: str
                Street Name
            zip: str
                5 digit zip code
            kwargs:
                Any additional keyword arguments will be passed to
                the EveryAction API for matching.
        `Returns:`
            A person dict
        """

        return self._people_search(
            first_name=first_name,
            last_name=last_name,
            date_of_birth=date_of_birth,
            email=email,
            phone=phone,
            phone_type=phone_type,
            street_number=street_number,
            street_name=street_name,
            zip=zip,
            create=True,
            **kwargs,
        )

    def upsert_person_json(self, match_json):
        """
        Create or update a person record.

        .. note::
            Person find must include the following minimum combinations.

              - first_name, last_name, email
              - first_name, last_name, phone
              - first_name, last_name, zip5, date_of_birth
              - first_name, last_name, street_number, street_name, zip5
              - email_address

        .. note::
            A full list of possible values for the json, and its structure can be found
            `here <https://docs.ngpvan.com/reference/people#peoplefindorcreate>`_. `vanId` can
            be passed to ensure the correct record is updated.

        .. warning::
            This method can only be run on MyMembers, EveryAction, MyCampaign databases.

        `Args:`
            match_json: dict
                A dictionary of values to match against and save.
        `Returns:`
            A person dict
        """

        return self._people_search(match_json=match_json, create=True)

    def _people_search(
        self,
        id=None,
        id_type=None,
        first_name=None,
        last_name=None,
        date_of_birth=None,
        email: Union[str, List[Dict[str, Union[str, bool]]], None] = None,
        phone=None,
        phone_type="H",
        street_number=None,
        street_name=None,
        zip=None,
        match_json=None,
        create=False,
        **kwargs,
    ):
        # Internal method to hit the people find/create endpoints

        addressLine1 = None
        if street_name and street_number:
            addressLine1 = f"{street_number} {street_name}"

        # Check to see if a match map has been provided
        if not match_json:
            json = {"firstName": first_name, "lastName": last_name}

            # Will fail if empty dicts are provided, hence needed to add if exist
            if email:
                if isinstance(email, str):
                    json["emails"] = [{"email": email}]
                elif isinstance(email, list):
                    json["emails"] = email
                else:
                    raise ValueError(f"Unexpected data type for email argument: {type(email)}")
            if phone:  # To Do: Strip out non-integers from phone
                json["phones"] = [{"phoneNumber": phone, "phoneType": phone_type}]
            if date_of_birth:
                json["dateOfBirth"] = date_of_birth
            if zip or addressLine1:
                json["addresses"] = [{}]
                if zip:
                    json["addresses"][0]["zipOrPostalCode"] = zip
                if addressLine1:
                    json["addresses"][0]["addressLine1"] = addressLine1
        else:
            json = match_json
            if "vanId" in match_json:
                id = match_json["vanId"]

        if kwargs:
            json.update(kwargs)

        url = "people/"

        if id:
            if create:
                id_type = "" if id_type in ("vanid", None) else f"{id_type}:"
                url += id_type + str(id)
            else:
                return self.get_person(id, id_type=id_type)

        else:
            url += "find"

            if create:
                url += "OrCreate"
            else:
                # Ensure that the minimum combination of fields were passed
                json_flat = json_format.flatten_json(json)
                self._valid_search(**json_flat)

        return self.connection.post_request(url, json=json)

    def _valid_search(
        self,
        firstName=None,
        lastName=None,
        email=None,
        phoneNumber=None,
        dateOfBirth=None,
        addressLine1=None,
        zipOrPostalCode=None,
        **kwargs,
    ):
        # Internal method to check if a search is valid, kwargs are ignored

        if (
            None in [firstName, lastName, email]
            and None in [firstName, lastName, phoneNumber]
            and None in [firstName, lastName, zipOrPostalCode, dateOfBirth]
            and None in [firstName, lastName, addressLine1, zipOrPostalCode]
            and None in [email]
        ):
            raise ValueError(
                """
                             Person find must include the following minimum
                             combinations to conduct a search.
                                - first_name, last_name, email
                                - first_name, last_name, phone
                                - first_name, last_name, zip, dob
                                - first_name, last_name, street_number, street_name, zip
                                - email
                            """
            )

        return True

    def get_person(
        self,
        id,
        id_type="vanid",
        expand_fields=None,
    ):
        """
        Returns a single person record using their VANID or external id.

        `Args:`
            id: str
                A valid id
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``. Defaults to ``vanid``.
            expand_fields: list
                A list of fields for which to include data. If a field is omitted,
                ``None`` will be returned for that field. Can be ``contribution_history``,
                ``addresses``, ``phones``, ``emails``, ``codes``, ``custom_fields``,
                ``external_ids``, ``preferences``, ``recorded_addresses``,
                ``reported_demographics``, ``suppressions``, ``cases``, ``custom_properties``,
                ``districts``, ``election_records``, ``membership_statuses``, ``notes``,
                ``organization_roles``, ``scores``, ``disclosure_field_values``.
        `Returns:`
            A person dict
        """

        # Change end point based on id type
        if expand_fields is None:
            expand_fields = [
                "contribution_history",
                "addresses",
                "phones",
                "emails",
                "codes",
                "custom_fields",
                "external_ids",
                "preferences",
                "recorded_addresses",
                "reported_demographics",
                "suppressions",
                "cases",
                "custom_properties",
                "districts",
                "election_records",
                "membership_statuses",
                "notes",
                "organization_roles",
                "disclosure_field_values",
            ]
        url = "people/"

        id_type = "" if id_type in ("vanid", None) else f"{id_type}:"
        url += id_type + str(id)

        # Removing the fields that are not returned in MyVoters
        NOT_IN_MYVOTERS = ["codes", "contribution_history", "organization_roles"]

        if self.connection.db_code == 0:
            expand_fields = [v for v in expand_fields if v not in NOT_IN_MYVOTERS]

        expand_fields = ",".join([json_format.arg_format(f) for f in expand_fields])

        logger.info(f"Getting person with {id_type or 'vanid'} of {id} at url {url}")
        return self.connection.get_request(url, params={"$expand": expand_fields})

    def delete_person(self, vanid):
        """
        Suppress the given VANID in databases where contact records can be suppressed.

        `Args:`
            vanid: str
                The person's VAN ID.
        `Returns:`
            Success or error.
        """
        url = f"people/{vanid}"
        r = self.connection.delete_request(url)
        logger.info(f"Van ID {vanid} suppressed.")
        return r

    def apply_canvass_result(
        self,
        id,
        result_code_id,
        id_type="vanid",
        contact_type_id=None,
        input_type_id=None,
        date_canvassed=None,
        phone=None,
    ):
        """
        Apply a canvass result to a person. Use this end point for attempts that do not
        result in a survey response or an activist code (e.g. Not Home).

        `Args:`
            id: str
                A valid person id
            result_code_id : int
                Specifies the result code of the attempt. Valid ids can be found
                by using the :meth:`get_canvass_responses_result_codes`
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
            contact_type_id : int
                `Optional`; A valid contact type id
            input_type_id : int
                `Optional`; Defaults to 11 (API Input)
            date_canvassed : str
                `Optional`; ISO 8601 formatted date. Defaults to todays date
            phone: str
                `Optional`; Phone number of any type (Work, Cell, Home)
        `Returns:`
            ``None``
        """

        logger.info(f"Applying result code {result_code_id} to {id_type} {id}.")
        self.apply_response(
            id,
            None,
            id_type=id_type,
            contact_type_id=contact_type_id,
            input_type_id=input_type_id,
            date_canvassed=date_canvassed,
            result_code_id=result_code_id,
            phone=phone,
        )

    def toggle_volunteer_action(
        self,
        id,
        volunteer_activity_id,
        action,
        id_type="vanid",
        result_code_id=None,
        contact_type_id=None,
        input_type_id=None,
        date_canvassed=None,
    ):
        """
        Apply or remove a volunteer action to or from a person.

        `Args:`
            id: str
                A valid person id
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
            volunteer_activity_id: int
                A valid volunteer activity id
            action: str
                Either 'apply' or 'remove'
            result_code_id : int
                `Optional`; Specifies the result code of the response. If
                not included,responses must be specified. Conversely, if
                responses are specified, result_code_id must be null. Valid ids
                can be found by using the :meth:`get_canvass_responses_result_codes`
            contact_type_id: int
                `Optional`; A valid contact type id
            input_type_id: int
                `Optional`; Defaults to 11 (API Input)
            date_canvassed: str
                `Optional`; ISO 8601 formatted date. Defaults to todays date

        ** NOT IMPLEMENTED **
        """

        """
        response = {"volunteerActivityId": volunteer_activity_id,
                    "action": self._action_parse(action),
                    "type": "VolunteerActivity"}

        logger.info(f'{action} volunteer activity {volunteer_activity_id} to {id_type} {id}')
        self.apply_response(id, response, id_type, contact_type_id, input_type_id, date_canvassed,
                            result_code_id)
        """

    def apply_response(
        self,
        id,
        response,
        id_type="vanid",
        contact_type_id=None,
        input_type_id=None,
        date_canvassed=None,
        result_code_id=None,
        omit_contact=False,
        phone=None,
        campaignId=None,
        skip_matching=False,
    ):
        """
        Apply responses such as survey questions, activist codes, and volunteer actions
        to a person record. This method allows you apply multiple responses (e.g. two survey
        questions) at the same time. It is a low level method that requires that you
        conform to the VAN API `response object
        format <https://docs.ngpvan.com/reference/canvass-responses>`_.

        `Args:`
            id: str
                A valid person id
            response: dict
                A list of dicts with each dict containing a valid action.
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
            result_code_id : int
                `Optional`; Specifies the result code of the response. If
                not included,responses must be specified. Conversely, if
                responses are specified, result_code_id must be null. Valid ids
                can be found by using the :meth:`get_canvass_responses_result_codes`
            contact_type_id : int
                `Optional`; A valid contact type id
            input_type_id : int
                `Optional`; Defaults to 11 (API Input)
            date_canvassed : str
                `Optional`; ISO 8601 formatted date. Defaults to todays date
            responses : list or dict
                The responses to apply.
            omit_contact: boolean
                Omit adding contact history to the response. This is particularly
                useful when adding activist codes that are not based on contact
                attempts.
            phone: str
                `Optional`; Phone number of any type (Work, Cell, Home)
            campaignId: int
                `Optional`; a valid Campaign ID.
            skip_matching: boolean
                `Optional`; if set to true, skips matching/de-duping of contact history. Defaults to a null value, aka false.
        `Returns:`
            ``True`` if successful

        .. code-block:: python

            response = [{"activistCodeId": 18917,
                         "action": "Apply",
                         "type": "ActivistCode"},
                        {"surveyQuestionId": 109149,
                         "surveyResponseId": 465468,
                         "type": "SurveyResponse"}
                        ]
            van.apply_response(5222, response)
        """

        # Set url based on id_type
        if id_type == "vanid":
            url = f"people/{id}/canvassResponses"
        else:
            url = f"people/{id_type}:{id}/canvassResponses"

        json = {
            "canvassContext": {
                "contactTypeId": contact_type_id,
                "inputTypeId": input_type_id,
                "dateCanvassed": date_canvassed,
                "omitActivistCodeContactHistory": omit_contact,
                "campaignId": campaignId,
                "skipMatching": skip_matching,
            },
            "resultCodeId": result_code_id,
        }

        if phone:
            json["canvassContext"]["phone"] = {
                "dialingPrefix": "1",
                "phoneNumber": phone,
            }

        if response:
            json["responses"] = response

        if result_code_id is not None and response is not None:
            raise ValueError("Both result_code_id and responses cannot be specified.")

        if isinstance(response, dict):
            json["responses"] = [response]

        if result_code_id is not None and response is not None:
            raise ValueError("Both result_code_id and responses cannot be specified.")

        return self.connection.post_request(url, json=json)

    def create_relationship(self, vanid_1, vanid_2, relationship_id):
        """
        Create a relationship between two individuals

        `Args:`
            vanid_1 : int
                The vanid of the primary individual; aka the node
            vanid_2 : int
                The vanid of the secondary individual; the spoke
            relationship_id : int
                The relationship id indicating the type of relationship
        `Returns:`
            ``None``
        """

        json = {"relationshipId": relationship_id, "vanId": vanid_2}

        self.connection.post_request(f"people/{vanid_1}/relationships", json=json)
        logger.info(f"Relationship {vanid_1} to {vanid_2} created.")

    def apply_person_code(self, id, code_id, id_type="vanid"):
        """
        Apply a code to a person.

        `Args:`
            id: str
                A valid person id.
            code_id: int
                A valid code id.
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
        `Returns:`
            ``None``
        """

        # Set url based on id_type
        if id_type == "vanid":
            url = f"people/{id}/codes"
        else:
            url = f"people/{id_type}:{id}/codes"

        json = {"codeId": code_id}

        self.connection.post_request(url, json=json)
        logger.info(f"Code {code_id} applied to person id {id}.")

    def merge_contacts(self, primary_vanid, source_vanid):
        """
        Merges two contacts in EveryAction. The source contact record will be
        deleted as part of the merge and its data will be moved into the primary
        contact record. In cases where fields conflict between the two contacts
        and we can't keep both values, such as if the contacts have different
        first names, the primary contact record's data will be retained. For
        more information see the
        `VAN API documentation here <https://docs.ngpvan.com/reference/peoplevanidmergeinto>`_

        `Args:`
            primary_vanid: str
                The VANID of the primary contact record.
            source_vanid: str
                The VANID of the source contact record.
        `Returns:`
            The VANID of the primary contact record.
        """

        url = f"people/{source_vanid}/mergeInto"
        json = {"vanId": primary_vanid}

        r = self.connection.put_request(url, json=json)
        return r



"""NGPVAN Canvass Responses Endpoints"""

import logging

from parsons.etl.table import Table

logger = logging.getLogger(__name__)


class CanvassResponses(object):
    def __init__(self, van_connection):
        self.connection = van_connection

    def get_canvass_responses_contact_types(self):
        """
        Get canvass response contact types.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request("canvassResponses/contactTypes"))
        logger.info(f"Found {tbl.num_rows} canvass response contact types.")
        return tbl

    def get_canvass_responses_input_types(self):
        """
        Get canvass response input types.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request("canvassResponses/inputTypes"))
        logger.info(f"Found {tbl.num_rows} canvass response input types.")
        return tbl

    def get_canvass_responses_result_codes(self):
        """
        Get canvass response result codes.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request("canvassResponses/resultCodes"))
        logger.info(f"Found {tbl.num_rows} canvass response result codes.")
        return tbl
